open Ptree
open Types
open Env
open Logic
open Logic_decl
open Cpn_options
open Cpn
open Cpn_check

(*s Processing of a program. *)

(*let is_token_type t =
  match t with
    PTexternal (_, id) -> if (String.compare (Ident.string id) "token") = 0 then true else false
  | _ -> false
*)
let rec logic_type_contains_token l = 
  match l with
    [e] -> (is_token_type e)
  | e::l' -> (is_token_type e) or (logic_type_contains_token l')
  | _ -> false

let rec add_arrow id lt r =
  let r_ty = r.c_result_type in
  match r_ty with
    PureType PTbool ->
      begin
	(if_debug_3 Format.eprintf "* Arrow %a is an invariant@." Ident.print id);
	Cpn.add_inv id (r.c_post)
        
      end
  | Arrow _ -> 
      begin
	(if_debug_3 Format.eprintf "* Arrow %a is a rule@." Ident.print id);
        add_arrow_rule id [] lt r  (* test if it is a rule and add it *)
      end
  | _ -> ()

and add_arrow_rule id lid lt r =
  let new_lid = lid@(token_id lt) 
  in
  match r.c_result_type  with
    Arrow (ltp,rp) -> add_arrow_rule id new_lid ltp rp
  | PureType PTunit -> 
	Cpn.add_rule id new_lid r.c_pre r.c_post
  | _ -> ()
	
and token_id lt =
  try
    match lt with
      [e] -> 
	let ty_id = (snd e) in
	let str_tid = (match ty_id with 
	  PureType (PTexternal ([],ty_id)) ->  (Ident.string ty_id) 
	| _ -> raise Exit
		      )
	in 
	if (String.compare str_tid "token") = 0
	then [(fst e)]
	else raise Exit
    | _ -> raise Exit
  with Exit -> []

let interp_cpn_decl ?(_prelude=false) d = 
  (* let lab = Label.empty in *)
  match d with 
    | Program (loc, id, p) ->
	(if_debug_3 Format.eprintf "* Program %a is not dealt@." Ident.print id)

    | Parameter (loc, ext, ids, v) -> 
	Main.interp_decl ~_prelude d;
	let add_cpn id =
	  let t = lookup_global id in
	  match t with
	    PureType _ -> ()
	  | Ref _ ->
	      begin
		(if_debug_3 Format.eprintf "* Parameter %a is a global variable@." Ident.print id);
		Cpn.add_globals id
	      end
	  | Arrow (lt,r) -> (* rule or invariant or something else *)
	      add_arrow id lt r
	in
	List.iter add_cpn ids

    | Exception (loc, id, v) ->
	Main.interp_decl ~_prelude d

    | Logic (loc, ext, ids, t0) ->
	Main.interp_decl ~_prelude d;
	let add_cpn id =
	  let vmap,t = find_global_logic id in
	  match t with
	    Predicate l -> 
	      if logic_type_contains_token l 
	      then
		if (List.length l)=1 then
		  begin
		    (if_debug_3 Format.eprintf "* Logic %a is a place@." Ident.print id);
		    Cpn.add_place id
		  end
		else
		  () (* TODO: verify that is equality of tokens *)
	      else ()
	  | Function (l, _) -> (* TODO: collect coloring symbols *)
	      if (logic_type_contains_token l) &&
		 (List.length l)=1
	      then
		begin
		  (if_debug_3 Format.eprintf "* Logic %a is a coloring symbol@." Ident.print id);
		  Cpn.add_color id
		end
	      else () (* TODO: verify that only colors use tokens *)
	in
	List.iter add_cpn ids

    | Predicate_def (loc, id, pl, p) -> 
	Main.interp_decl ~_prelude d

    | Function_def (loc, id, pl, ty, e) ->
	if_debug_3 Format.eprintf "* Function %a is not dealt@." Ident.print id

    | Inductive_def(loc,id,t,indcases) ->
	if_debug_3 Format.eprintf "* Inductive definition %a is not dealt@." Ident.print id

    | Axiom (loc, id, p) -> (* TODO *)
	Main.interp_decl ~_prelude d

    | Goal (loc, id, p) -> 
	Main.interp_decl ~_prelude d

    | TypeDecl (loc, ext, vl, id) ->
	Main.interp_decl ~_prelude d

(*s Processing of a channel / a file. *)
  
let deal_cpn_channel parsef cin =
  let p = parsef cin in
  if not Options.parse_only then List.iter interp_cpn_decl p

let deal_cpn_file f =
  Main.reset ();
  let cin = open_in f in 
  deal_cpn_channel (Main.why_parser f) cin;
  close_in cin;
  let fwe = Filename.chop_extension f in
  if not (Main.single_file ()) then Main.output (Options.out_file fwe)

let check_cpn () =
  begin
    (* add to invariants informations about places concerned *)
    Format.printf "- compute special form\n"; 
    Cpn.decorate_inv ();
    (* check invariant on given rules *)
    Format.printf "- check invariants using post\n"; 
    Cpn_check.check_invariant true;  (* using post, TODO: put option for post/pre *)
    ()
  end

let main_cpn () =
  let t0 = Unix.times () in
  (* Main.load_prelude (); *)
  if files = [] then begin
    deal_cpn_channel (Main.why_parser "standard input") stdin;
    check_cpn ();
    Main.output (Options.out_file "out")
  end else begin
    List.iter deal_cpn_file files;
    check_cpn ();
    if Main.single_file () then 
      let lf = Filename.chop_extension (Misc.last files) in
      Main.output (Options.out_file lf)  
  end;
  (Cpn.print_cpn stdout);
  if show_time then
    let t1 = Unix.times () in
    Format.printf "Why execution time : %3.2f@." 
      (t1.Unix.tms_utime -. t0.Unix.tms_utime)

