
open Ident
open Logic

type ty_cpn_rule = {
    name : Ident.t;
    leftside : Ident.t Ident.map; (* token:place *)
    rightside : Ident.t Ident.map; (* token:place *)
    cond : predicate
  }

type ty_inv = {
    subinv : predicate;  (* subinvariant, i.e., a disjunction *)
    splaces : Ident.set; (* set of places concerned by the invariant*)   
    isempty : bool       (* flag true iff places is empty and false iff places is the full set *)
  }

type ty_cpn_model =  {
    places : Ident.set;  (* logic *)
    globals : Ident.set; (* parameters *)
    colors : Ident.set;  (* logic *)
    rules : ty_cpn_rule list;
    inv : ty_inv list    (* conjunction of the elements of the list (subinvariants) *)
  }

(* Global variable storing the system *)
let cpn_model : ty_cpn_model ref = ref {
  places = Idset.empty;
  globals = Idset.empty;
  colors = Idset.empty;
  rules = [];
  inv = []
}

(**
 * Build the system
 *)
let add_place id = 
  if (String.compare (Ident.string id) "bot")<>0 
  then 
    cpn_model := {
      places = Idset.add id (!cpn_model).places;
      globals = (!cpn_model).globals;
      colors = (!cpn_model).colors;
      rules = (!cpn_model).rules;
      inv = (!cpn_model).inv
    }

let add_color id = 
  cpn_model := {
    places = (!cpn_model).places;
    globals = (!cpn_model).globals;
    colors = Idset.add id (!cpn_model).colors;
    rules = (!cpn_model).rules;
    inv = (!cpn_model).inv
  }

let add_globals id = 
  cpn_model := {
    places = (!cpn_model).places;
    globals = Idset.add id (!cpn_model).globals;
    colors = (!cpn_model).colors;
    rules = (!cpn_model).rules;
    inv = (!cpn_model).inv
  }

exception Bad_rule of string

let is_token_type t =
  match t with
    PTexternal (_, id) -> 
      if (String.compare (Ident.string id) "token") = 0 then true else false
  | _ -> false                            

let rec add_rule id lt c_pre c_post_list = 
  try
    cpn_model := {
      places = (!cpn_model).places;
      globals = Idset.add id (!cpn_model).globals;
      colors = (!cpn_model).colors;
      rules = (!cpn_model).rules@[(build_rule id lt c_pre c_post_list)];
      inv = (!cpn_model).inv
    }
  with Bad_rule _ -> ()

and build_rule id lt c_pre_list c_post_opt =
  let c_pre = (match c_pre_list with [] -> Ptrue | e::_ -> e)
  and c_post = (match c_post_opt with None -> Ptrue | Some(a,_) -> a)
  in
  let ls_pre, rs_pre, pre_new = build_constraint lt c_pre
  and rs_post, ls_post, post_new = build_constraint lt c_post
  in
  if (is_left_side ls_pre ls_post) &&
    (is_right_side rs_pre rs_post) 
  then 
    { name = id;
      leftside = ls_pre;
      rightside = rs_post;
      cond = Pand(false,false, (* TODO: init correctly these params *)
		   pre_new, post_new)
    }
  else
    raise  (Bad_rule  "error")

and build_constraint lt c =
  match c with
    Pand (_,_,lc,rc) -> 
      let pl_l,bot_l,cl = build_constraint lt lc
      and pl_r,bot_r,cr = build_constraint lt rc
      in 
      let pl = add_token_maps pl_l pl_r
      and bot = add_token_maps bot_l bot_r
      in
      (pl, bot, build_constraint_and cl cr)
      
  | Papp (pl,argl,_) -> 
      if (Idset.mem pl (!cpn_model).places || (String.compare (Ident.string pl) "bot")=0 ) then
	let tid = get_token lt argl
	in
	if (String.compare (Ident.string pl) "bot")=0 then	  
          Idmap.empty, (Idmap.add tid pl Idmap.empty), Ptrue (*remove bot(t) in the rule condition *)
         
	else
	  (Idmap.add tid pl Idmap.empty), Idmap.empty, Ptrue
      else
	(Idmap.empty, Idmap.empty, c)

  | _ -> (Idmap.empty, Idmap.empty,c)

and add_token_maps ll lr =
  let add_token_map t pl m =
    (if Idmap.mem t m then raise (Bad_rule "token with two places")
     else Idmap.add t pl m)
  in
  Idmap.fold add_token_map ll lr

and build_constraint_and cl cr =
  match cl,cr with
    (Ptrue,Ptrue) -> Ptrue
  | (Ptrue,_) -> cr
  | (_,Ptrue) -> cl
  | (_,_) -> Pand(false,false,cl,cr)

and get_token lt args =
  match args with
    [e] -> (
      match e with
	Tvar tid -> 
	  if List.mem tid lt then
	    tid
	  else
	    raise (Bad_rule "token not defined")
      | _ -> raise (Bad_rule "bad place predicate parameter")
	    )
  | _ -> raise (Bad_rule "bad place predicate")

and is_left_side lpre lpost = true (* TODO *)
and is_right_side lpre lpost = true (* TODO *)


exception Bad_inv of string
       
let rec add_inv c_post_opt = 
  let c_post = (match c_post_opt with None -> [] | Some(a,_) -> (add_inv_special a))
  and add_places i = { subinv = i; splaces = Idset.empty; isempty = false }
  in
  cpn_model := {
    places = (!cpn_model).places;
    globals = (!cpn_model).globals;
    colors = (!cpn_model).colors;
    rules = (!cpn_model).rules;
    inv = ((!cpn_model).inv)@(List.map add_places c_post)
  }


(**  
 * Check the special form of invariants.
 *)
(** - conjunction of atoms **)
and add_inv_special a =
  match a with
    Pand (_,_,p1,p2) -> 
	let linv1 = add_inv_special p1
        in  (match linv1 with 
	      [Ptrue] -> (add_inv_special p2)
	    | [Pfalse] -> linv1
            | _ -> let linv2 = add_inv_special p2 
		   in (match linv2 with
                         [Ptrue] -> linv1
                       | [Pfalse] -> linv2
                       | _ -> linv1@linv2
		   )
            )

  | Ptrue -> [Ptrue]
  | Pfalse -> [Pfalse]
  | _ -> [(add_inv_special_or a)]

(** - disjunction of (possibly negated) quantified formula **)
and add_inv_special_or a = 
  match a with 
    Ptrue -> Ptrue
  | Pfalse -> Pfalse
  | Por (p1,p2) -> 
	let p1' = add_inv_special_or p1
        in  (match p1' with 
	      Pfalse -> (add_inv_special_or p2)
	    | Ptrue -> p1'
            | _ -> let p2' = add_inv_special_or p2 
		   in (match p2' with
                         Pfalse -> p1'
                       | Ptrue -> Ptrue
                       | _ -> Por(p1',p2')
		   )
            )
  | Pnot p -> 
      let p' = add_inv_special_or p
      in 
      (match p' with 
	Pfalse -> Ptrue
      | Ptrue -> Pfalse
      | Pnot p1 -> p1
      | _ -> (Pnot p')
      )
  | Exists _ -> begin (add_inv_special_sigma1 a); a end
  | Forall _ -> begin (add_inv_special_pi1 a); a end
  | _ -> raise(Bad_inv "inv is not a conjunction of formulas in Sigma1 and Pi1")

(** - list of quantified variables in Sigma1 **)
and add_inv_special_sigma1 a =
  match a with
   (* check the type of the quantified variable *)
  | Exists (_,_,_,p) -> (add_inv_special_sigma1 p)

  | Forall (_,_,_,ty,_,p) ->
      if (is_token_type ty) (* formula is no more in Sigma1, check Sigma0 *)
      then raise (Bad_inv "the subinvariant is not in Sigma1") 
      else  (add_inv_special_sigma0 p)

  | _ -> add_inv_special_quantifier_free a 

(** - list of quantified variables in Pi1 **)
and add_inv_special_pi1 a =
  match a with
   (* check the type of the quantified variable *)
  | Forall (_,_,_,_,_,p) -> add_inv_special_pi1 p
		
  | Exists (_,_,ty,p) -> 
      if (is_token_type ty) (* formula is no more in Pi1, check Sigma0 *)
      then   (add_inv_special_sigma0 p)
      else   raise (Bad_inv "the subinvariant is not in Pi1") 

  | _ -> add_inv_special_quantifier_free a 

(** - list of quantified variables in Sigma0 **)
and add_inv_special_sigma0 a =
 match a with
  | Forall (_,_,_,ty,_,p) -> 
      if (is_token_type ty)
      then raise (Bad_inv "the subformula is not in Sigma0")
      else   add_inv_special_sigma0 p

  | Exists (_,_,ty,p) -> 
      if(is_token_type ty)
      then  raise (Bad_inv "the subformula is not in Sigma0")
      else   add_inv_special_sigma0 p
		

  | _ -> add_inv_special_quantifier_free a 

(** - a quantifier free formula **)
and  add_inv_special_quantifier_free a =
 match a with

  | Exists _ | Forall _ | Forallb _ ->
	raise (Bad_inv "the subformula is not quantified free")
  
  | Pvar _ | Papp _ | Ptrue | Pfalse | Pfpi _ -> ()
  | Pimplies (_,p1,p2) | Pif (_,p1,p2) | Pand (_,_,p1,p2) 
  | Por (p1,p2) | Piff (p1,p2) -> 
      begin 
	(add_inv_special_quantifier_free p1); 
	(add_inv_special_quantifier_free p2)
      end
  | Pnot p | Pnamed (_,p) -> (add_inv_special_quantifier_free p) 



 
(**
 * Decorate with places used by the invariant of the model.
 *)
let rec decorate_inv () =
  let new_inv = List.map decorate_subinv ((!cpn_model).inv)
  in
  cpn_model := {
    places = (!cpn_model).places;
    globals = (!cpn_model).globals;
    colors = (!cpn_model).colors;
    rules = (!cpn_model).rules;
    inv = new_inv
  }


(**
 * Decorate with places used by a subinvarint.
 * @param inv  structure subinvariant, set of places, flag
 * @return     structure subinvarint (not changed), 
 *             computed set of places,
 *             and flag
 *)
and decorate_subinv inv  =
  let p, f = decorate_subinv_aux inv.subinv
  in  { subinv = inv.subinv; splaces = p; isempty = f }

and decorate_subinv_aux inv =
  match inv with
    Ptrue | Pfalse -> Idset.empty,true
  | Por (p1,p2) -> 
      let pl1,f1 = decorate_subinv_aux p1 in 
        if (Idset.is_empty pl1) && not(f1)
	then (* full set of places, return *)
	  pl1, f1
	else (
	  let pl2, f2 = decorate_subinv_aux p2 in
          if (Idset.is_empty pl2) && not(f2)
	  then (* full set of places, return *)
	    pl2, f2
	  else (
	    let pl12 = Idset.union pl1 pl2 in
	    if (Idset.is_empty pl12)
	    then
	      Idset.empty, true
	    else if (Idset.equal (!cpn_model.places) pl12)
	    then
	      Idset.empty, false
	    else
	      pl12, false
	   )
	 )
  | Pnot p -> decorate_subinv_aux p 

  | Exists _  -> let tp = decorate_subinv_exist_quant inv [] in (* Sigma1 formula *)
    places_of tp 
  | Forall _  -> let tp = decorate_subinv_univ_quant  inv [] in (* Pi1 formula *)
    places_of tp;
  | _ -> (* NOT ALLOWED by the simplified form of invariants *)
      Idset.empty, true

(**
 * Computes an overapproximation of the set of places concerned 
 * by the subinvariant.
 * @param inv    invariant in Pi1
 * @param tp     list of pairs (token identifier, set of places associated)
 * @return       updated list of pairs tp
 *)
and decorate_subinv_univ_quant inv tp = 
  match inv with
   Forall (_,_,t,ty,triggs,p) -> if (is_token_type ty)
                                 then let pl = decorate_token_triggers t triggs in 
                                 (* collect places for this token from triggers *)
			         decorate_subinv_univ_quant p tp@[(t,pl)]
				 else
			         decorate_subinv_univ_quant p tp
  
  | Exists (_,_,ty,p) ->  (* we are in Pi1, so we have quantifiers over colors
                           * which shall be skiped to find the free quantified formula *)
                           let psi = skip_sigma0_quantifiers p in
                           (* collect the rest of the possible places *)
                           decorate_subinv_univ_quant psi tp
                          
  | Pimplies (_,p1,p2) -> decorate_subinv_and p1 tp

  | _ -> tp


(**
 * Computes an overapproximation of the set of places concerned 
 * by the subinvariant.
 * @param inv    invariant in Sigma1
 * @param tp     list of pairs (token identifier, set of places associated)
 * @return       list of places concerned by the invariant with flag
 *)
and decorate_subinv_exist_quant inv tp = 
  match inv with
  | Exists (_,t,ty,p) -> if (is_token_type ty)
                         then decorate_subinv_exist_quant p (tp@[(t,Idset.empty)])
			 else decorate_subinv_exist_quant p tp

  | Forall (_,_,_,ty,_,p) -> (* we are in Sigma1  so we have quantifiers over colors
                              * which shall be skiped to find the free quantified formula *)
                             let psi=skip_sigma0_quantifiers p in
			     decorate_subinv_exist_quant psi tp 

  | Pand (_,_,_,_) -> decorate_subinv_and inv tp
  
  | _ -> tp


and decorate_subinv_and inv tp =
  match inv with
    Pand (_,_,p1,p2) -> 
      let tp1 = (decorate_subinv_and p1 tp) in
      (decorate_subinv_and p2 tp1)
  | Papp(pl,[(Tvar v)],_) ->
      if (Idset.mem pl (!cpn_model.places))
      then (* is a place *)
	try 
          let spl = List.assoc v tp 
	  and ntp = List.remove_assoc v tp in
	  let nspl = Idset.union spl (Idset.singleton pl) in
	    (v,nspl)::ntp        
	 with Not_found -> tp (* Error in typing of Why? *)

      else
	 tp

  | _ -> tp

(**
 * Collect the set of places in the triggers for a given token var
 * @param t    token var
 * @param lt   list of triggers concerning the token
 * @return     set of places in the triggers concerning the token
 *)
and decorate_token_triggers t lt =
 match lt with
    [] -> Idset.empty 
  | e::lt' -> let pl = decorate_token_triggers t lt' in
		Idset.union (decorate_token_trigger t e) pl

and decorate_token_trigger t trig = 
 match trig with 
   [PPat(Papp(pl,[(Tvar v)],_))] -> 
		if (String.compare (Ident.string t) (Ident.string v)) = 0 
		then (Idset.singleton pl)
		else Idset.empty
 | _ -> Idset.empty 
	 
     
and skip_sigma0_quantifiers p =
 match p with
  | Forall (_,_,_,_,_,p1) -> skip_sigma0_quantifiers p1
  | Exists (_,_,_,p1) -> skip_sigma0_quantifiers p1
  | _ -> p (* free quantified formula *)


and places_of tp = 
  let union_places pl t_p = Idset.union pl (snd t_p) in
  let s = List.fold_left union_places Idset.empty tp in
  if (Idset.is_empty s)  (* empty set means here quantified token but not located *)
  then s, true
  else 
  if (Idset.equal (!cpn_model.places) s)
  then
    Idset.empty, false
  else
    s, false
 

(**
 * Printing the system.
 *)
let rec print_cpn outchan =
  begin
    (Format.print_string "\nPlaces: ");
    print_id_set (!cpn_model).places;
    (Format.print_string "\nGlobals: ");
    print_id_set (!cpn_model).globals;
    (Format.print_string "\nColors: ");
    print_id_set (!cpn_model).colors;
    (Format.print_string "\nRules: ");
    print_rules (!cpn_model).rules;
    (Format.print_string "\nInvariant: ");
    print_invs (!cpn_model).inv
  end

and print_id id = 
  (*TODO if debug then
    (Ident.dbprint Format.std_formatter id)
  else *)
    (Ident.print Format.std_formatter id)
    
and print_id_set idset = 
  begin
  (Idset.iter print_id idset);
  (Format.print_string " ")
  end

and print_rules rl =
  List.iter print_rule rl

and print_rule r =
  begin
    (Format.print_string "\n\n\n( ");
    (print_id r.name);
    (Format.print_string ": ");
    (Idmap.iter print_token_place r.leftside);
    (Format.print_string " -> ");
    (Idmap.iter print_token_place r.rightside);
    (Format.print_string " : \n ");
    (Util.print_predicate Format.std_formatter r.cond);
    (Format.print_string ")\n ")
  end

and print_token_place tid pid =
  begin
    (Format.print_string " ");
    (print_id tid);
    (Format.print_string ":");
    (print_id pid);
    (Format.print_string " ")
  end

and print_invs invl =
  begin
    (Format.print_string " of length ");
    (Format.print_int (List.length invl));
    (Format.print_string "\n");
    (List.iter print_inv invl)
  end

and print_inv inv =
  begin   
    (Format.print_string "\n\n && [places: "); 
    (if (Idset.is_empty inv.splaces)
     then if not(inv.isempty)
         then (Format.print_string "all")
         else (Format.print_string "none")
     else
      (print_id_set inv.splaces)
    )
    ;
    (Format.print_string "] ");
    (Util.print_predicate Format.std_formatter inv.subinv)
    end

