
open Options
open Ident
open Logic
open Types
open Main
open Cpn
open Cpn_prepost
open Smtlib
open Util


exception Invariant_not_true of string
exception Unexpected_SAT of string
exception ErrorPNF of string
exception ErrorSAT of string

type quant_ty = { quant: bool; (* false for exist, true for forall *)
		  id: Ident.t; 
		  idp: Idset.t; 
		  ty: Logic.pure_type }

let print_quant_ty lq =
  List.iter (fun q -> begin
    (if q.quant then Format.print_string "all " else Format.print_string "ex ");
    (Ident.dbprint Format.std_formatter q.id);
    (Format.print_string " in { ");
    (Idset.iter (fun id -> begin
      (Ident.dbprint Format.std_formatter id);
      (Format.print_string " ")
    end ) q.idp);
    (Format.print_string "} . ");
  end) lq

type allmap_ty = { allvar: Ident.t;
                   exvars: Ident.t array }


module I = struct type t_ = predicate type t = t_ let compare = compare end

module Predset = Set.Make(I)

(** pset is a conjunct of predicate and this function add p to it 
 *)
let add_pred_to_predset isstop p pset =
  let return_or_stop b pset =
    (if isstop then raise Not_found
    else pset)
  in
  if (Predset.mem Pfalse pset) then
    return_or_stop isstop pset
  else 
    (match p with
      Pfalse -> return_or_stop isstop (Predset.singleton Pfalse)
    | Ptrue -> pset
    | _ -> (Predset.add p pset)
    )

(** counter needed for temporal files *)
let cnt_tmp = ref 0

(**
 * Check invariant given in the model on each rule.
 * Stop to check when find a rule not checked by raising Invariant_not_true.
 *
 * @param p  true if post is used, false for pre
 * @return   unit
 *)
let rec check_invariant p =
  List.iter (check_invariant_rule p) (!Cpn.cpn_model.rules)

and check_invariant_rule p r =
  begin
    (* compute invariants touched or not by the rule *)
    let inv_r,inv_not_r = split_invariant r (!Cpn.cpn_model.inv) in

    (* compute post/pre on invariants touched by the rule *)
    let p_inv_r = List.map (fun inv -> inv.subinv) inv_r in
    let p_inv_not_r = List.map (fun inv -> inv.subinv) inv_not_r in
    let cset,lp_inv_r = Cpn_prepost.image_of_rule p r p_inv_r in
 
    (* generate a list of |inv_r| unsat problems to check the rule *)
    List.iter (check_invariant_unsat r.rightside cset lp_inv_r p_inv_not_r) inv_r 
  end


(**************** Split I = {I_rule} and {I_not_rule} ***************************)
(**
 * Split the list given by @param linv on two lists depending
 * on places used in @param r.
 * @param r    rule
 * @param linv list of subinvariants
 * @return     pair of lists of invariants (l_r, l_not_r)
 *             l_r refers to places of @param r, l_not_r does not. 
 *)
and split_invariant r linv =
  let union_map t p pset = Idset.add p pset in
  let r_places = Idset.union (Idmap.fold union_map r.leftside (Idset.empty)) 
                             (Idmap.fold union_map r.rightside (Idset.empty)) in 
  let linv_r, l_inv_not_r = List.fold_left (split_subinv r_places) ([],[]) linv in
  begin
    (Format.print_string "\n**********************************************");
    (Format.print_string (String.concat " " 
			  ["\n** Rule"; 
			   (Ident.string r.name);
			   "(r="]));
    (Format.print_int (List.length linv_r));
    (Format.print_string ", !r=");
    (Format.print_int (List.length l_inv_not_r));
    (Format.print_string ")");
    (Format.print_flush ());
    (linv_r, l_inv_not_r)
  end

and split_subinv rset pinv inv =
 if (Idset.is_empty inv.splaces) && not(inv.isempty) (* inv concerns all places *)
 then ((fst pinv)@[inv], (snd pinv))
 else let mem_set s e = (Idset.mem e s) in
      let is_r = (Idset.exists (mem_set rset) inv.splaces) in
      if is_r then
        ((fst pinv)@[inv], (snd pinv))
      else
        ((fst pinv), (snd pinv)@[inv])



(**************** SAT FORMULAS ***************************)

(**
 * Generate a sat problem for the formula
 *  exists yset. exists cset. /\ p_inv_r /\ inv_not_r /\ not (inv_r_j)
 * If the problem is not unsat, raise exception Invariant_not_true. 
 * 
 * @param ymap      existentially quantified token variables
 * @param cset      existentially quantified color variables
 * @param p_inv_r   list of formula representing the pre/post image 
 *                  of invariants
 * @param inv_not_r list of invariants not concerned by the image
 * @param inv_r_j   invariant concerned by the image
 * @return          unit or raise exception
 *)
and check_invariant_unsat ymap cset p_inv_r inv_not_r inv_r_j =
  begin
    (Format.print_string "\n\t+not ");
    (Format.print_string (Ident.string inv_r_j.invname));
    (Format.print_string ":");
    (Format.print_flush ());
    (** build prenex form for p_inv_r and (and inv_not_r) and not(inv_r_ *)
    (* - rename quantified variables in inv_r_j *)
    let p_inv_r_j = rename_bound_vars Idmap.empty inv_r_j.subinv in 
    
    (* - put in prenex normal form *)  
    let lq1,lp1 = prenex_form_preds p_inv_r 
    and lq2,lp2 = prenex_form_preds inv_not_r
    and lq3,lp3 = prenex_form_preds [Pnot(p_inv_r_j)] in 
    let lp123 = Cpn_prepost.union_preds [lp1;lp2;lp3] in
    let lcq1 = Idset.fold quant_add_color cset lq1 in
    let lycq1 = Idmap.fold quant_add_token ymap lcq1 in
    let lq = quant_combine [lycq1; lq2; lq3] in

       (* apply the sat procedure to obtain a pure color predicate psi *)
       (* generate file for psi and call dp *)
       (* analyse the result of the dp and return or raise exception *)
    (** call the sat procedure on CML, result expected unsat (false) *)
    check_pred_sat_sigma2 false lq lp123
  end
(***************RENAME*******************************)

(** put new timestamps on quantified variables of phi **)
(*returns the predicate phi with the substitution*)
 (*
    * goes through the formula and for each quantifier variable x do
    * let n = Ident.bound x in
    * and continue with new substitution x -> n
    *)
and rename_bound_vars sigma phi =  
  
  match phi with
    Pvar _  -> phi  (* Warning: predicate variables not used in CML *)
      
  | Pif _ -> raise (ErrorSAT "If predicate: not yet implemented!")

  | Pand (wp,sym,p1,p2) ->  let p'1 =  rename_bound_vars sigma p1 in 
                             let p'2 =  rename_bound_vars sigma p2 in 
                               Pand (wp,sym,p'1,p'2)

	
  | Por (p1, p2) -> let p'1 =  rename_bound_vars sigma p1 in 
                      let p'2 =  rename_bound_vars sigma p2 in 
                        Por (p'1,p'2)


  | Piff (p1,p2) ->  let p'1 =  rename_bound_vars sigma p1 in 
                      let p'2 =  rename_bound_vars sigma p2 in 
                        Piff (p'1,p'2)


  | Pnot p -> let p' = rename_bound_vars sigma p in 
                 Pnot(p')
 
  | Pfpi _  -> 
      raise (ErrorSAT "Fpi formulas: not yet implemented!") 

  | Pnamed _ ->
      raise (ErrorSAT "Named formulas: not yet implemented!")

  | Ptrue ->  Ptrue

  | Pfalse -> Pfalse

  | Pimplies (wp,p1,p2) -> let p'1 =  rename_bound_vars sigma p1 in 
                            let p'2 =  rename_bound_vars sigma p2 in 
                               Pimplies (wp, p'1,p'2)

              
  | Forall (wp,id,id',ty,trig,p) ->(*new quantified var ==> new substitution*)
                                   let newid' = Ident.bound id' in (*generate the new name*) 
                                       let p'= rename_bound_vars ( Idmap.add id' newid' sigma ) p in (*add a new substitution*)
                                        let trig_sigma = rename_var_triggers trig  (Idmap.add id' newid' sigma) in (* make the substitution in the triggers*)
                                             Forall (wp,id,newid',ty,trig_sigma,p')
                                                                                              
  | Forallb _ ->
      raise (ErrorSAT "Forallb formulas: not yet implemented!")

  | Exists (id,id',ty,p) ->(*new quantified var ==> new substitution*)
                            let newid' = Ident.bound id' in (*generate the new name*) 
                              let p'= rename_bound_vars ( Idmap.add id' newid' sigma ) p in (*add a new substitution*)
                                Exists (id,newid',ty,p')


  | Papp(p,lterm,lty) ->(*the substitution is done in the list of terms*) 
                        let lterm_sigma = rename_var_lterms lterm sigma in
                          Papp(p,lterm_sigma,lty) 


and rename_var_triggers ltrig sigma =
 match ltrig with
    [] -> []
  | e::lt' -> let ltrig'= rename_var_triggers lt' sigma in
                   [rename_var_trig e sigma] @ ltrig'               

and rename_var_trig trig sigma = 
 match trig with 
   [PPat(Papp(pl,[(Tvar v)],i))] -> (* apply substitution sigma *)
     (try
       [PPat(Papp(pl,[(Tvar(Idmap.find v sigma))],i))] 
     with Not_found ->trig)
   | _ -> trig


and rename_var_lterms lterm sigma =
 match lterm with
    [] -> []
  | e::lt' -> let lterm'= rename_var_lterms lt' sigma in
                   [rename_var_term e sigma] @ lterm'               

and rename_var_term term sigma =
  match term with
    Tconst _ -> term
  | Tderef x -> (* TODO *) Tderef x

  | Tvar x -> 
      (* apply substitution sigma *)
      (try
        Tvar (Idmap.find x sigma)
      with Not_found ->
        Tvar x)

   | Tapp (f,lt,i) -> Tapp (f,rename_var_lterms lt sigma ,i)
                     


(** add existential quantifier on color cvar before lq **)
and quant_add_color cvar lq =
  ({quant = false; id = cvar; idp = Idset.empty; ty = PTint(* TODO *)})::lq

(** add existential quantifier on token cvar before lq **)
and quant_add_token yid pid lq =
  { quant = false; id = yid; idp = (Idset.singleton pid); ty = (PTexternal ([], Ident.create "token"))}::lq

(************************PRENEX NORMAL FORM**************************) 
(** compute pnf for each element of lp **)
(** return quantifiers and token-quantifier free list of formula *)
(*lp is a list of predicate belonging to B(Sigma1)*)
and prenex_form_preds lp =
  (* DOING *)
  begin
  (* Printf.fprintf stdout "\n\t(Prenex "; *)
  (* 1st step: build the list of pairs [(lq1,p1),....(lqn,pn)] *)
  let llqp = List.fold_left (prenex_form_preds_acc) [] lp in
   
  (* 2nd step: build the pair ([lq1,....lqn] , [p1,.....,pn]) *)
  let llq,llp = List.fold_left (accumulate_pair) ([],[]) llqp  in 

  (* 3rd step: make a quantifier combination, build (lq,[p1,.....,pn]):PNF *)
  let lq = quant_combine llq in
  begin
 (* begin debug
    (print_quant_ty lq); 
    List.iter (fun p -> begin
      (Format.print_string "\n && ");
      (Util.print_predicate Format.std_formatter p)
    end) llp;
    Format.print_string "\n\t)\n";
    end debug *)
    lq,llp
  end
  end


    (***** PNF: 1st step: build the list of pairs [(lq1,p1),....(lqn,pn)]******)

(* @param  lpnf:cummulate the list of pairs [(lq1,p1),....(lqn,pn)]*)
and prenex_form_preds_acc lpnf p = 
  (* compute prenex form *)
  let lqi,pi = prenex_form_pred [] false p in
  begin 
(* begin debug 
  *  Format.print_string "\n\t(Prenex_form input\n";
  *  Util.print_predicate Format.std_formatter p ;
  *  Format.print_string "\n\t)\n\t(Prenex_form output \n";
  *  (print_quant_ty lq);
  *  Util.print_predicate Format.std_formatter pi;
  *  Format.print_string "\n\t)\n";
  * end debug *)
    lpnf @ [(lqi,pi)]
  end
(*we compute first the list of quantified var and the predicate modified after removing the triggers for each predicate of the list, we do it in one "passage"*)
(*returns: list of quantified var in order, predicate without triggers*)
(*ltq: to cummulate the list of quantified var, initially empty*)
and prenex_form_pred tp isneg pred = 
  match pred with
    Pvar _  -> [],pred  (* Warning: predicate variables not used in CML *)
      
  | Pif _ -> raise (ErrorSAT "If predicate: not yet implemented!")

  | Pand (wp,sym,p1,p2) -> let ltq1,p'1 = prenex_form_pred tp isneg p1  
                           and ltq2,p'2 = prenex_form_pred tp isneg p2 in 
			   let ltqq = (quant_combine [ltq1;ltq2]) in
                               (match isneg, p'1, p'2 with
                                  false, Ptrue, Ptrue -> [], Ptrue
                               |  false, _, Pfalse | false, Pfalse, _ -> ltqq, Pfalse
                               |  false, Ptrue, _ -> ltqq, p'2
                               |  false, _, Ptrue -> ltqq, p'1
                               |  false, _, _ -> ltqq, Pand(wp,sym,p'1,p'2)
			       |  true, Pfalse, Pfalse -> ltqq, Pfalse
			       |  true, Ptrue, _ |  true, _, Ptrue -> [], Ptrue
			       |  true, Pfalse, _ -> ltqq, p'2
			       |  true, _, Pfalse -> ltqq, p'1
                               |  true, _, _ -> ltqq, Por(p'1,p'2)
                                ) 
	
  | Por (p1, p2) -> let ltq1,p'1 = prenex_form_pred tp isneg p1
                    and ltq2,p'2 = prenex_form_pred tp isneg p2 in
                    let ltqq = (quant_combine [ltq1;ltq2]) in
                               (match isneg, p'1, p'2 with
                                  true, Ptrue, Ptrue -> [], Ptrue
                               |  true, _, Pfalse |  true, Pfalse, _ -> ltqq, Pfalse
                               |  true, Ptrue, _ -> ltqq, p'2
                               |  true, _, Ptrue -> ltqq, p'1
                               |  true, _, _ -> ltqq, Pand(false, true,p'1,p'2)
                               |  false, Pfalse, Pfalse -> ltqq, Pfalse
                               |  false, Ptrue, _ |  false, _, Ptrue -> [], Ptrue
                               |  false, Pfalse, _ -> ltqq, p'2
                               |  false, _, Pfalse -> ltqq, p'1
                               |  false, _, _ -> ltqq, Por(p'1,p'2)
                                )
                                
  | Piff (p1,p2) -> let ltq1,p'1 = prenex_form_pred tp (if isneg then true else false) p1
                    and ltq2,p'2 = prenex_form_pred tp false p2 in 
                    (quant_combine [ltq1;ltq2]), Piff (p'1,p'2)
                          

  | Pnot p -> prenex_form_pred tp (not(isneg)) p

 
  | Pfpi _  -> 
      raise (ErrorSAT "Fpi formulas: not yet implemented!") 

  | Pnamed _ ->
      raise (ErrorSAT "Named formulas: not yet implemented!")

  | Ptrue ->  [], (if isneg then Pfalse else Ptrue)

  | Pfalse -> [], (if isneg then Ptrue else Pfalse)

  | Papp(p,lterm,lty) -> [], (if isneg then Pnot(Papp(p,lterm,lty)) else Papp(p,lterm,lty))

  | Pimplies (wp,p1,p2) ->  let ltq1,p'1 = prenex_form_pred tp (not(isneg)) p1 
                            and ltq2,p'2 = prenex_form_pred tp isneg p2 in
                            let ltqq = (quant_combine [ltq1;ltq2]) in
			    (match isneg, p'1, p'2 with
				false, Ptrue, _ 
			     |  false, _, Ptrue -> [], Ptrue
			     |  false, Pfalse, _ -> ltqq, p'2
			     |  false, _, Pfalse -> ltqq, p'1
			     |  false, Pnot(p), _ -> ltqq, Pimplies(wp, p, p'2)
			     |  false, _, _ -> ltqq, Pimplies(wp, Pnot(p'1), p'2)
			     |  true, Pfalse, _ 
			     |  true, _, Pfalse -> ltqq, Pfalse
			     |  true, Ptrue, _-> ltqq, p'2
			     |  true, _, Ptrue -> ltqq, p'1
			     |  true, _, _ -> ltqq, Pand(false,true, p'1, p'2)
			    )

  | Exists (id,id',ty,p) -> 
      (* begin optimization: localize exists tokens not localized *)
      let tpnew = (match tp with 
	[] -> Cpn.decorate_subinv_exist_quant pred tp
      | _ -> tp) 
      in 
      let ltq',p' = prenex_form_pred tpnew isneg p in
      let pset = (try List.assoc id' tpnew with Not_found -> Idset.empty) in
      ({quant=(if isneg then true else false);id=id';idp=pset;ty=ty})::ltq',p'
          
  | Forall (wp,id,id',ty,trig,p) ->
      (* new quantified var ==> add to the list,
         remove the forall predicate 
         and add predicate to p to replace the triggers *)          
      (* TODO : collect places in p for id' *)
      let ltq',p' = prenex_form_pred tp isneg p  
      and ptrig = build_predicate_triggers trig in
      let ltqq = ({quant= (if isneg then false else true);
		   id=id';
		   idp=(Cpn.decorate_token_triggers id' trig);
		   ty=ty})::ltq' in
      (match isneg, ptrig, p' with
	false, Ptrue, _ -> ltqq, p'
      | false, _, Pfalse -> ltqq, Pnot(ptrig)
      | false, _, Ptrue -> [], Ptrue
      | false, _, _ -> ltqq, Pimplies(wp,ptrig,p')
      | true, Ptrue, Ptrue -> [], Ptrue
      | true, Ptrue, _ -> ltqq, p'
      | true, _, Pfalse -> ltqq, Pfalse
      | true, _, Ptrue -> ltqq, ptrig
      | true, _, _ -> ltqq, Pand(wp,true,ptrig,p')
      )

  | Forallb _ ->
      raise (ErrorSAT "Forallb formulas: not yet implemented!")


and build_predicate_triggers ltrig = 
  match ltrig with
    [] -> Ptrue
  | [p] -> (* forall u :token [p(u)].Phi <=> forall u:token p(u)=> Phi *)
           build_pred_trigg p
  | _ ->  (*forall u :token [p(u)|q(u)] <=> forall u:token p(u) v q(u) => Phi*) 
          build_predicate_or_triggers ltrig 


and build_predicate_or_triggers ltrigg =
   match ltrigg with
     [e] -> (build_pred_trigg e)
   | e::lt' -> let ptrig'= build_predicate_or_triggers lt' in
                  Por(build_pred_trigg e,ptrig')
   | _ -> (* not reached *) Pfalse              

and build_pred_trigg trig =
 match trig with 

   [PPat(p)] -> p 
  | _ -> Pfalse (*Error: not possible within the CPN system, triggers are only used to express places ==> we can't have TPat(terms) *)

and negate_quantifiers lq =
  List.fold_left (negate_quantifiers_acc) [] lq

and negate_quantifiers_acc not_lq q = (*q has the form :(int,ident t,pure type) *)
(*reverse the quantifiers*)
  not_lq@[{quant=(not q.quant);id=q.id;idp=q.idp;ty=q.ty}]

       (*****PNF: 2nd step: build the pair ([lq1,....lqn] , [p1,.....,pn])*******)
and accumulate_pair pairqp qp =
   (fst pairqp)@ [(fst qp)],(snd pairqp)@[(snd qp)] 



	(****** 3rd step: make a quantifier combination, build (lq,[p1,.....,pn]):PNF ******)
(** combine list of quantifiers in @param llq to obtain a unique list **)
and quant_combine llq =

  let llist_rev lq llqacc =  (List.rev lq)::llqacc in
  let llqrev = List.fold_right llist_rev llq [] in (* reverse each list of llq *)
  (*q_comb = list of combined quantifiers*)
  (*llq_not_sigma0 = list of lists of quantified vars with sigma0 variables removed*)
  let q_comb,llq_not_sigma0 =  combine_sigma0 llqrev in

  (*llq_not_pi1 = list of lists of quantified vars with pi1 variables removed*)
  let q_comb1,llq_not_pi1 = combine_pi1 q_comb llq_not_sigma0 in 

  (*llq_not_sigma2 = list of lists of quantified vars with sigma2 variables removed*)
  let q_comb2,llq_not_sigma2 = combine_sigma2 q_comb1 llq_not_pi1 in 

  match llq_not_sigma2 with 
    [] ->  q_comb2
  | _ ->  raise (ErrorPNF "the formulas are in pi3")

 
(**
 * combines variables of sigma0 formulas
 *
 * @param q_comb   accumulate the quantifiers combined, initially empty
 * @param lq       list of qualtified variables for one formula, the list is reversed
 * @return      list of combined quantifiers,list of quantified variables for this formula with the sigma0 quantifiers removed
 *)

and combine_sigma0 llq =
 List.fold_left (combine_sigma0_acc) ([],[]) llq
 
and combine_sigma0_acc (q_comb,llq_not_sigma0) lq =
 match lq with
  [] -> (q_comb,llq_not_sigma0)
 | e::lq1 -> 
     if ((Cpn.is_token_type (e.ty))== false)(*it's a color variable => add it to q_comb, remove it from lq*)
     then combine_sigma0_acc (e::q_comb,llq_not_sigma0) lq1

     else (*it's a token variable => we re no more in sigma0, we're in pi1*) 
       (q_comb,llq_not_sigma0@[lq])



and combine_pi1 q_comb llq =
  List.fold_left (combine_pi1_acc) (q_comb,[]) llq

and combine_pi1_acc (q_comb,llq_not_pi1) lq =
(*parcourir the list and check if it's a quantified universally token*)
  match lq with
    [] -> (q_comb,llq_not_pi1)
  | e::lq1 -> 
      if (e.quant == true) 
          (*it's a token/color var quantified universally=>add it to q_comb, remove it from lq*)  
      then combine_pi1_acc (e::q_comb,llq_not_pi1) lq1
       
      else (*it's an existential variable => we are no more in pi1, we're in sigma2*) 
        (q_comb,llq_not_pi1@[lq])


and combine_sigma2 q_comb llq =
  List.fold_left (combine_sigma2_acc) (q_comb,[]) llq


and combine_sigma2_acc (q_comb,llq_not_sigma2) lq =
(*parcourir the list and check if it's a quantified existentially variable*)
  match lq with
    [] -> (q_comb,llq_not_sigma2)
  | e::lq1 ->
      if (e.quant == false)
          (*it's a token/color var quantified existentially=>add it to q_comb, remove it from lq*)
      then combine_sigma2_acc (e::q_comb,llq_not_sigma2) lq1
      else (*it's an universally variable => we are no more in sigma2, we're in pi3 *) 
        (q_comb,llq_not_sigma2@[lq]) (*TODO: when lq is empty*)


(** return if the Sigma2 formula (lq. lp) is 
 *  - sat for issat true, 
 *  - unsat otherwise
 *  raise exception otherwise 
 *)
and check_pred_sat_sigma2 issat lq lp =
  begin
(* begin debug 
  (Format.print_string "\n(Check unsat input \n ");
  (print_quant_ty lq); 
  (Format.print_string ", [ ");
  (Format.print_int (List.length lp));
  (Format.print_string "]\n");
  (List.iter (fun p -> begin
    (Format.print_string "\n && ");
    (Util.print_predicate Format.std_formatter p)
  end) lp);
  (Format.print_string "\n)");
  (Format.print_string "\n(Check unsat output \n ");
end debug *)

  (* step 1: from lq build two mappings 
   *     exallmap : p -> (set of token vars quantified ex and in p,
   *                      set of tokens quantified all and in p)
   *     loc : v -> set of places where v may be located
   *  if a var ex is not localized add all places as localization
   *)
  let exallmap, loc = List.fold_left (build_ex_all_mapping) 
                                     (Idmap.empty,Idmap.empty) lq 
  in

  (* step 2: build the mapping giving for each all variable the set
   *    of ex variables which can replace it wrt to localization.
   *)
  let sigma = Idmap.fold build_matching exallmap Idmap.empty
  in

  (* step 3: split sigma into 
   *    sigma1 : only all vars such that |sigma(allv)| <= 1
   *      i.e., at most one choice is possible for this variable
   *    sigma2 : variables all vars with more than one choice
   *)
  let sigma1, sigma2 = Idmap.fold filter_simple_matching sigma 
                                      (Idmap.empty,Idmap.empty)
  in
(* begin debug 
   let _ = begin
   (Format.print_string "\n sigma1 : ");
   print_sigma sigma1;
   (Format.print_string "\n sigma2 : ");
   print_sigma sigma2
   end
   end
   in
   end debug *)

  (** step 4: apply sigma1 on lp to simplify. 
   *)
  let lp1 = (List.map (apply_loc_substitution sigma1 loc) lp) in
  let lp1' = Cpn_prepost.union_preds [lp1] in
(* begin debug
   let _ = begin
   (Format.print_string "\n after sigma1: [ ");
   (Format.print_int (List.length lp1'));
   (Format.print_string "]\n");
   List.iter (fun p -> begin
   (Format.print_string "\n && ");
   (Util.print_predicate Format.std_formatter p)
   end) lp1';
   end
   in
   end debug *)

  (** step 5: generate on the fly all the substitutions possible from sigma2
   *          and the set of conjuncts obtained from formula lp1'
   *)
  let sp2 = reduce_sigma2_loc sigma2 loc lp1' in 
(* begin debug
   let _ = begin
   (Format.print_string "\n after sigma2: [ ");
   (Format.print_int (Predset.cardinal sp2));
   (Format.print_string "]\n");
   Predset.iter (fun p -> begin
   (Format.print_string "\n && ");
   (Util.print_predicate Format.std_formatter p)
   end) sp2;
   end
   in
   end debug *)

  (** step 6: eliminate universal quantifiers on tokens from lq and vmap
   *)
  let lq', loc' = remove_allquant_vars lq loc 
  in

  (** step 7: generate sigma1 sat problems for each possible localization
   *          of existentially quantified variables
   *)
  check_pred_sat_sigma1_loc issat lq' loc' sp2
end
    
 
  
and build_ex_all_mapping pmap q =
  if (Cpn.is_token_type q.ty)
  then 
    let pset = (if (Idset.is_empty q.idp) then (!Cpn.cpn_model).places 
                else q.idp) 
    in
    (Idset.fold (build_mapping (q.quant) q.id) pset (fst pmap)),
    (Idmap.add q.id pset (snd pmap))
  else
    pmap


and build_mapping quant idt pl map = (* pl -> (ex_var_set, all_var_set) *)
    let oldpair = (try Idmap.find pl map 
                   with Not_found -> (Idset.empty,Idset.empty)) 
    in
    let newpair = (if quant then (fst oldpair, Idset.add idt (snd oldpair)) 
                   else (Idset.add idt (fst oldpair), snd oldpair)) 
    in
    Idmap.add pl newpair map

and build_matching pl exall sigma1 =
    let size_ex = Idset.cardinal (fst exall) in
    let newres = (if size_ex = 0 then Idset.empty else (fst exall)) in
    let sigma1 = Idset.fold (build_matching_all newres) (snd exall) sigma1 in
    sigma1

and build_matching_all newres allid sigma1 =
    let oldres = (try Idmap.find allid sigma1 
                  with Not_found -> Idset.empty) in 
    let nres = Idset.union oldres newres in 
    Idmap.add allid nres sigma1

(** sigma1 maps a variable in the set all to a variable in the set ex 
 *  if and only if the set ex has size <= 1
 *)
and filter_simple_matching idall exset sigma12 =
    if (Idset.cardinal exset) > 1 then 
      ((fst sigma12), (Idmap.add idall exset (snd sigma12)))
    else 
      ((Idmap.add idall exset (fst sigma12)), (snd sigma12))

(** apply a substitution which may be not total to a formula pred
 *  the localization of tokens is in vmap
 *)
and apply_loc_substitution sigma1 vmap pred =
  (* begin debug
    (Format.print_string "\n apply_subst sigma 1 input: ");
    (Util.print_predicate Format.std_formatter pred);
    let p = end debug *) 
  (match pred with
    Pvar _  -> pred  (* Warning: predicate variables not used in CML *)
      
  | Papp(p,lterm,lty) -> 
      (* case 1: equality of tokens t1=t2 *)
      if ((String.compare (Ident.string p) "eq_token") == 0)
      then (* lterm = [(Tvar x);(Tvar y)] *)
	(
        let t1 = apply_loc_subst_term sigma1 vmap (List.hd lterm)
	and t2 = apply_loc_subst_term sigma1 vmap (List.hd (List.tl lterm)) 
	in (* extract identifiers of tokens in terms *) 
	match (t1,t2) with
	  (Tvar x, Tvar y) -> 
	    if x = y then (* TODO: verify structural equality *)
	      Ptrue
	    else 
	      let x_pset = (try (Idmap.find x vmap) with Not_found -> Idset.empty)
	      and y_pset = (try (Idmap.find y vmap) with Not_found -> Idset.empty)
	      in if Idset.is_empty (Idset.inter x_pset y_pset) then
		Pfalse (* not equality with disjoint locations *)
	      else 
		Papp(p,[t1;t2],lty)

	| (_,_) -> (* Error: not possible after type checking *)
	    raise (ErrorSAT "t1=t2 formula: result unexpected")
	)
      else (* case 2: place predicate p(t) *)
	if (Idset.mem p (!Cpn.cpn_model).places) 
	then (* lterm = [(Tvar t)] *)
	  match lterm with 
	    [(Tvar tid)] -> 
	      let issubst,xset = (try true, Idmap.find tid sigma1 with Not_found -> false, Idset.empty) in
	      if issubst && (Idset.is_empty xset) then
		Pfalse (* i.e., no token can be in any place *) 
	      else if issubst then
                let lterm' = List.fold_left (apply_loc_subst_term_acc sigma1 vmap) [] lterm in
		  apply_loc_papp vmap (Papp(p,lterm',lty))
              else (* not substitution but maybe localization *)
		  apply_loc_papp vmap pred
          | _ -> (* Warning: not allowed after type-checking *)
	      raise (ErrorSAT "p(t) formula: result unexpected")
      else (* case 3: predicate on colors *)
        let lterm' = List.fold_left (apply_loc_subst_term_acc sigma1 vmap) [] lterm
        in  Papp(p,lterm',lty)

  | Ptrue -> Ptrue

  | Pfalse -> Pfalse

  | Pimplies (wp,p1,p2) ->
      let p1' = apply_loc_substitution sigma1 vmap p1 in
      (match p1' with
         Pfalse -> Ptrue
       | Ptrue -> apply_loc_substitution sigma1 vmap p2
       | _ ->
           let p2' = apply_loc_substitution sigma1 vmap p2 in
           (match p2' with
             Ptrue -> Ptrue
           | _ -> Pimplies (wp, p1', p2')
           )
       )

  | Pand (wp,sym,p1,p2) ->
      let p1' = apply_loc_substitution sigma1 vmap p1 in
      (match p1' with
         Pfalse -> Pfalse
       | Ptrue -> apply_loc_substitution sigma1 vmap p2
       | _ ->
           let p2' = apply_loc_substitution sigma1 vmap p2 in
           (match p2' with
             Pfalse -> Pfalse
           | Ptrue -> p1'
           | _ -> Pand (wp, sym, p1', p2')
           )
       )

  | Por (p1, p2) ->
      let p1' = apply_loc_substitution sigma1 vmap p1 in
      (match p1' with
        Ptrue -> Ptrue
      | Pfalse -> apply_loc_substitution sigma1 vmap p2
      | _ ->
          let p2' = apply_loc_substitution sigma1 vmap p2 in
          (match p2' with
            Ptrue -> Ptrue
          | Pfalse -> p1'
          | _ -> Por (p1', p2')
          )
      )

  | Piff (p1,p2) ->
      let p1' = apply_loc_substitution sigma1 vmap p1 
      and p2' = apply_loc_substitution sigma1 vmap p2 in
      (match p1', p2' with
        Ptrue, Ptrue
      | Pfalse, Pfalse -> Ptrue
      | Ptrue, Pfalse
      | Pfalse, Ptrue -> Pfalse
            (* TODO: more simplifications *)
      | _,_ ->  Piff (p1', p2')
      )

  | Pnot p ->
      let p' = apply_loc_substitution sigma1 vmap p in
      (match p' with
        Ptrue -> Pfalse
      | Pfalse -> Ptrue
      | _ -> Pnot p'
      )

  | Forall (wp,id,id',ty,trig,p) ->
      if (Cpn.is_token_type ty) then
	raise (ErrorSAT "Quantifier on token unexpected in quantifier free formula")
      else
      let p' = apply_loc_substitution sigma1 vmap p in
        (match p' with
          Pfalse | Ptrue -> p' (* we are with quantification over colors *)
        | _ -> Forall(wp,id,id',ty,trig,p')
        )

  | Exists (id,id',ty,p) ->
      if (Cpn.is_token_type ty) then
	raise (ErrorSAT "Quantifier on token unexpected in quantifier free formula")
      else
      let p' = apply_loc_substitution sigma1 vmap p in
        (match p' with
          Pfalse | Ptrue -> p'
        | _ -> Exists(id,id',ty,p')
        )

  | _ -> raise (ErrorSAT "Formulas not yet implemented!")
	)
   (*  in begin debug
    (Format.print_string "\n apply_subst sigma 1 output: ");
    (Util.print_predicate Format.std_formatter p);
      p
    end
  end debug *)


and apply_loc_subst_term sigma vmap term =
  match term with
    Tconst _ -> term
  | Tvar tid ->
      let issubst,xset = (try true, Idmap.find tid sigma with Not_found -> false, Idset.empty) in
      if not(issubst) then
	term
      else if Idset.is_empty xset then
	raise (ErrorSAT "substitution in term by an empty set")
      else (* there is a set with only one element *)
        let set2list x xlist = x::xlist in
	let xlist = (Idset.fold set2list xset []) in
          Tvar (List.hd xlist)

  | Tderef x -> (* TODO *) term

  | Tapp (f,lt,i) ->
      (* test that f is a color symbol *)
      if (Idset.mem f (!cpn_model.colors))
      then (* lt = [(Tvar t)] with t token by typing *)
        (let t1 = apply_loc_subst_term sigma vmap (List.hd lt) in Tapp (f,[t1], i))
      else (* function on color logic *)
        let lt' = List.fold_left (apply_loc_subst_term_acc sigma vmap) [] lt
        in  Tapp(f,lt',i)

and apply_loc_subst_term_acc sigma vmap lt_acc t =
	lt_acc@[(apply_loc_subst_term sigma vmap t)]


and apply_loc_papp vmap pred =
  match pred with
    Papp(p,[Tvar tid],lty) ->
      (let pset = (try Idmap.find tid vmap with Not_found -> Idset.empty) in
	match (Idset.cardinal pset), (Idset.mem p pset) with
          1, true -> Ptrue
        | _, false -> Pfalse
        | _, _ -> pred
      )
  | _ -> pred

and apply_loc_acc loc pred predset =
  let pred' = apply_loc loc pred in
  add_pred_to_predset true pred' predset

and apply_loc loc pred =
  (match pred with
    Pvar _  -> pred (* Warning: predicate variables not used in CML *)
      
  | Papp(p,lterm,lty) -> 
      if (Idset.mem p (!Cpn.cpn_model).places) 
      then (* lterm = [(Tvar t)] *)
	(match lterm with 
	  [(Tvar tid)] -> 
	    (let pl = (try Idmap.find tid loc with Not_found -> raise (ErrorSAT "localization not found!")) in
	    if (String.compare (Ident.string p) (Ident.string pl))=0 then
	      Ptrue
	    else
	      Pfalse
	    )
        | _ -> (* Warning: not allowed after type-checking *)
	      raise (ErrorSAT "p(t) formula: result unexpected")
	)
      else
	pred

  | Ptrue -> Ptrue

  | Pfalse -> Pfalse

  | Pimplies (wp,p1,p2) ->
      let p1' = apply_loc loc p1 in
      (match p1' with
         Pfalse -> Ptrue
       | Ptrue -> apply_loc loc p2
       | _ ->
           let p2' = apply_loc loc p2 in
           (match p2' with
             Ptrue -> Ptrue
           | _ -> Pimplies (wp, p1', p2')
           )
       )

  | Pand (wp,sym,p1,p2) ->
      let p1' = apply_loc loc p1 in
      (match p1' with
         Pfalse -> Pfalse
       | Ptrue -> apply_loc loc p2
       | _ ->
           let p2' = apply_loc loc p2 in
           (match p2' with
             Pfalse -> Pfalse
           | Ptrue -> p1'
           | _ -> Pand (wp, sym, p1', p2')
           )
       )

  | Por (p1, p2) ->
      let p1' = apply_loc loc p1 in
      (match p1' with
        Ptrue -> Ptrue
      | Pfalse -> apply_loc loc p2
      | _ ->
          let p2' = apply_loc loc p2 in
          (match p2' with
            Ptrue -> Ptrue
          | Pfalse -> p1'
          | _ -> Por (p1', p2')
          )
      )

  | Piff (p1,p2) ->
      let p1' = apply_loc loc p1 
      and p2' = apply_loc loc p2 in
      (match p1', p2' with
        Ptrue, Ptrue
      | Pfalse, Pfalse -> Ptrue
      | Ptrue, Pfalse
      | Pfalse, Ptrue -> Pfalse
            (* TODO: more simplifications *)
      | _,_ ->  Piff (p1', p2')
      )

  | Pnot p ->
      let p' = apply_loc loc p in
      (match p' with
        Ptrue -> Pfalse
      | Pfalse -> Ptrue
      | _ -> Pnot p'
      )

  | Forall (wp,id,id',ty,trig,p) ->
      let p' = apply_loc loc p in
        (match p' with
          Pfalse | Ptrue -> p' (* we are with quantification over colors *)
        | _ -> Forall(wp,id,id',ty,trig,p')
        )

  | Exists (id,id',ty,p) ->
      let p' = apply_loc loc p in
        (match p' with
          Pfalse | Ptrue -> p'
        | _ -> Exists(id,id',ty,p')
        )

  | _ -> raise (ErrorSAT "Formulas not yet implemented!")
	)

(** compute the Sigma1 formula obtained by eliminating from lp the
 *  all vars given by sigma2.
 *)
and reduce_sigma2_loc sigma2 vmap lp =
  let pset = ref Predset.empty
  and sigma2array = map2array sigma2 in
  let sigma2size = (Array.length sigma2array) in
  begin
  (try (
  let rec choose_one_subst iy ix sigma1 =
    (if iy >= sigma2size then
       (* sigma1 contains a subst for all all vars *)
       let lpsigma1 = (List.map (apply_loc_substitution sigma1 vmap) lp) in
       begin
         (List.iter (fun p -> (pset := (add_pred_to_predset true p !pset))) lpsigma1);
(* begin debug
	 (Format.print_string "(choose_one_subst sigma1 ");
	 print_sigma sigma1;
	 (Format.print_string ")\n");
	 (Format.print_string "(choose_one_subst added \n");
	 List.iter (fun p -> begin
	   (Format.print_string "\n && ");
	   (Util.print_predicate Format.std_formatter p)
	 end) lpsigma1;
	 (Format.print_string "\n");
	 (Format.print_string "\nnew pset[");
	 (Format.print_int (Predset.cardinal !pset));
	 (Format.print_string "])\n");
   end debug *)
       end
     else let ymap = (Array.get sigma2array iy) in
       if ix >= (Array.length ymap.exvars) then
         (* end of loop for iy *)
         ()
       else begin
         (* generate next level of substitutions for the choice of ix for iy *)
         (choose_one_subst (iy+1) 0 (Idmap.add (ymap.allvar) (Idset.singleton (Array.get (ymap.exvars) ix)) sigma1));
	 (* generate next choice for this iy *)
	 (choose_one_subst iy (ix+1) sigma1)
       end)
   in
      choose_one_subst 0 0 Idmap.empty
   )
   with Not_found -> (pset := Predset.singleton Pfalse))
   ;
   !pset
   end
           

and map2array sigma2 =
  let sigma2list = Idmap.fold map2list sigma2 [] in
  Array.of_list sigma2list

and map2list y xset lacc =
  let xlist = set2list xset in
  let ymap = { allvar=y; exvars=(Array.of_list xlist) } in
  ymap::lacc

and set2list s =
  (Idset.fold (fun e lacc -> lacc@[e]) s [])

and print_sigma s = 
  begin
    (Format.print_string " [ ");
    (Idmap.iter (fun idall exset ->
      begin
	(Format.print_string " { ");
	(Ident.dbprint Format.std_formatter idall);
	(Format.print_string " -> ");
	(Idset.iter (fun id -> begin
	  (Ident.dbprint Format.std_formatter id);
	  (Format.print_string " ")
	end ) exset);
	(Format.print_string " } ")
      end
		) s);
    (Format.print_string "\n ]\n ");
    (Format.print_flush ())
  end

(** remove universally quantified token variables from lq and loc *)
and remove_allquant_vars lq loc =
  List.fold_left (remove_allquant_var) ([],loc) lq

and remove_allquant_var lqloc q =
  if q.quant && (Cpn.is_token_type q.ty) then
    ((fst lqloc), (Idmap.remove q.id (snd lqloc)))
  else
    (q::(fst lqloc), (snd lqloc))


(** Uses loc to generate sigma1 sat problems and call
 *  @see check_pred_sat_sigma1 to solve them accoridng to
 *  expected result issat.
 *)
and check_pred_sat_sigma1_loc issat lq loc predset = 
  let loc_array = map2array loc in
  let loc_size = (Array.length loc_array) in
  let rec choose_one_loc ix ip loc1 =
    (if ix >= loc_size then
      (* loc1 contains a localization for all ex vars *)
      try (
	let predset1 = (Predset.fold (apply_loc_acc loc1) predset Predset.empty) in
	check_pred_sat_sigma1 issat lq predset1
       ) with Not_found -> (* predset is false *)
	 (if issat then
	   raise (Invariant_not_true "")
	 else
	   (* so unsat, continue *) ()
	 )
    else let xmap = (Array.get loc_array ix) in
      if ip >= (Array.length xmap.exvars) then
        (* end of loop for ix *)
        ()
      else begin
        (* generate next level of substitutions for the choice of ip for ix *)
        (choose_one_loc (ix+1) 0 (Idmap.add (xmap.allvar) (Array.get (xmap.exvars) ip) loc1));
	(* generate next choice for this ix *)
	(choose_one_loc ix (ip+1) loc1)
      end)
  in
  choose_one_loc 0 0 Idmap.empty
  

(** Generate a query to the solver and monitor the result:
 *  - if result=sat and issat then return else raise Unexpected_SAT
 *  - if result=unsat and not(issat) then return else raise Unexpected_SAT
 *  - if result=unknown then raise Unexpected_SAT
 *)
and check_pred_sat_sigma1 issat lq predset =
   begin
   (* initialize the environment *)
   Smtlib.reset ();
    
   (* TODO: push decl for logic AUFLIA *)

   (* TODO: push decl for color domain type *)

   (* push decl for colors *)
   Idset.iter push_decl_color (!cpn_model).colors;

   (* push decl for globals TODO: remove rules *)
   (* Idset.iter push_decl_global (!cpn_model).globals; *)

   (* push decl for predicate *)
   (let p = build_smt_pred lq predset in
   push_decl_goal p)
   ;

   (* print to a file and call Yices *)
   let tmp_file = "tmp/cpn_sat"^(string_of_int !cnt_tmp) in
   begin
     Smtlib.output_file tmp_file;
     cnt_tmp := !cnt_tmp+1;
     let r = Calldp.yices (tmp_file^"_why.smt") () in
(*
     if issat then
       (match r with
	 Calldp.Valid _ -> raise (Unexpected_SAT "sat expected, unsat result")
       | Calldp.Invalid _ -> ()
       | Calldp.CannotDecide _ -> raise (Unexpected_SAT "sat expected, unknown result")
       | Calldp.ProverFailure (_,s) ->  raise (Unexpected_SAT ("sat expected, prover failure "^s))
       | _ ->  raise (Unexpected_SAT "sat expected, prover failure")
       )
     else
       (match r with
	 Calldp.Valid _ -> ()
       | Calldp.Invalid _ -> raise (Unexpected_SAT "unsat expected, sat result")
       | Calldp.CannotDecide _ -> raise (Unexpected_SAT "unsat expected, unknown result")
       | Calldp.ProverFailure (_,s) ->  raise (Unexpected_SAT ("unsat expected, prover failure "^s))
       | _ ->  raise (Unexpected_SAT "unsat expected, prover failure")
       )
*)
     (* begin debug *)
     (Format.print_string ("\n(yices on "^(tmp_file)^" returns "));
     (match r with
       Calldp.Valid _ -> (Format.print_string "unsat)")
     | Calldp.Invalid _ -> (Format.print_string "sat)")
     | Calldp.CannotDecide _ -> (Format.print_string "unknown)")
     | Calldp.ProverFailure (_,s) -> (Format.print_string (s^")"))
     | _ ->  raise (Unexpected_SAT "sat expected, prover failure")
     );
     (Format.print_flush ())
     (* end debug *)
   end
  end

and push_decl_color cid =
  let col_decl = Logic_decl.Dlogic(Loc.dummy_position, (Ident.string cid),
			(Env.empty_scheme (Logic.Function([PTint], PTint)))) (* TODO: change PTint to color domain *)
  in
	Smtlib.push_decl col_decl
  
and push_decl_global gid =
  let glob_decl = Logic_decl.Dlogic(Loc.dummy_position, (Ident.string gid),
			(Env.empty_scheme (Logic.Function([], PTint)))) (* TODO: change PTint to color domain *)
  in
    Smtlib.push_decl glob_decl

and push_decl_goal p =
  let goal_decl = Logic_decl.Dgoal(Loc.dummy_position, "cpn",
			(Env.empty_scheme ([], p)))
  in begin 
(* begin debug
    Format.print_string "(push_goal \n";
    Util.print_predicate Format.std_formatter p;
    Format.print_string ")\n";
   end debug *)
    Smtlib.push_decl goal_decl
  end

and build_smt_pred lq pset =
  begin 
  (* begin debug
    (Format.print_string "\n(added quantifiers [");
    (Format.print_int (List.length lq));
    (Format.print_string "])\n");
     end debug *)
   let p = Predset.fold build_smt_pred_acc pset Ptrue in
   let lq_p = List.fold_right build_smt_quant_pred_acc lq p in
     lq_p
  end

and build_smt_pred_acc pred pacc =
   let pred' = rename_token2int pred in
   match pred',pacc with
     _, Ptrue -> pred'
   | Ptrue, _ -> pacc
   | _, Pfalse -> Pfalse
   | Pfalse, _ -> Pfalse
   | _, _ -> Pand(true, false, pred', pacc)

and rename_token2int pred =
  (match pred with
    Pvar _  -> pred (* Warning: predicate variables not used in CML *)
      
  | Papp(p,lterm,lty) -> 
      if ((String.compare (Ident.string p) "eq_token") == 0)
      then 
        Papp(Ident.t_eq_int, lterm, lty)
      else
	pred

  | Ptrue -> Ptrue

  | Pfalse -> Pfalse

  | Pimplies (wp,p1,p2) ->
      let p1' = rename_token2int p1 in
      let p2' = rename_token2int p2 in
        Pimplies (wp, p1', p2')

  | Pand (wp,sym,p1,p2) ->
      let p1' = rename_token2int p1 in
      let p2' = rename_token2int p2 in
        Pand (wp, sym, p1', p2')

  | Por (p1, p2) ->
      let p1' = rename_token2int p1 in
      let p2' = rename_token2int p2 in
        Por (p1', p2')

  | Piff (p1,p2) ->
      let p1' = rename_token2int p1 in
      let p2' = rename_token2int p2 in
        Piff (p1', p2')

  | Pnot p ->
      let p' = rename_token2int p in
        Pnot p'

  | Forall (wp,id,id',ty,trig,p) ->
      let p' = rename_token2int p in
        Forall(wp,id,id',ty,trig,p')

  | Exists (id,id',ty,p) ->
      let p' = rename_token2int p in
        Exists(id,id',ty,p')

  | _ -> raise (ErrorSAT "Formulas not yet implemented!")
	)
   
and build_smt_quant_pred_acc q pred =
   let ty = (if (Cpn.is_token_type q.ty) then PTint else q.ty) in
   if (q.quant) then
     Forall(true, q.id, q.id, ty, [], pred)
   else
     Exists(q.id, q.id, ty, pred)

	
