
open Options
open Ident
open Logic
open Types
open Main
open Cpn
open Util


exception ErrorPrePost of string

(**************** Compute of POST and PRE ***************************)
(**
 * Computes post/pre image of formulas @param i1_in by rule @param r.
 *
 * @param p     true if post is used, false for pre
 * @param r     rule applied
 * @param i1_in list of formulas on which the image is computed
 * @return      list of formulas representing the image
 *)

let rec image_of_rule p r i1_in =
  if p then
    image_of_rule_post r i1_in
  else
    image_of_rule_pre r i1_in


(******* POST *******)
and image_of_rule_post r i1_in =

  (* apply ominus (r.leftside) on the list [r.cond]@i1_in
   *       and return 
   *  - the list of formulas transformed by ominus
   *  - the set of color variables used by resulting formulas above
   *)

  let lp, cset = ominus r.leftside ((r.cond)::i1_in) in

  (* apply oplus (r.rightside) on result of ominus
   *       and return a new list of predicates
   *)
  let lp' = oplus r.rightside lp 
  and lpy = Idmap.fold split_loc r.rightside []
  in 
  begin
(* begin debug 
    Printf.fprintf stdout "\n(Post_%s \n" (Ident.string r.name); 
    (Idmap.iter (fun t p -> begin
      (Format.print_string " ex ");
      (Ident.dbprint Format.std_formatter t);
      (Format.print_string ":token [");
      (Ident.dbprint Format.std_formatter p);
      (Format.print_string "]")
    end) r.rightside);
    (Idset.iter (fun c -> begin
      (Format.print_string " ex ");
      (Ident.dbprint Format.std_formatter c)
    end) cset);
    (List.iter (fun p -> begin
      (Format.print_string "\n && ");
      (Util.print_predicate Format.std_formatter p)
    end)
      (lpy@lp'));
    Printf.fprintf stdout "\n) \n";
 end debug *)
    (* build predicate (or predicate list) Exist quantified by r.rightside and cset *)
    cset, lpy@lp'
  end

and split_loc y p yp_acc = yp_acc@[Papp(p,[(Tvar y)],[])]



(******* PRE *******)
and image_of_rule_pre r i1_in = 
  (* TODO *) Idset.empty, [Ptrue]

(** Simplifications for conjunction list of predicates **)
and union_preds llp =
  union_preds_acc [Ptrue] llp

and union_preds_acc lr llp =
  (match llp with
    [] -> lr
  | lp::llp' -> let lr' = union_pred_acc lr lp in
    (match lr' with 
      [Pfalse] -> [Pfalse]
    | _ -> union_preds_acc lr' llp'
    )
  )
    
and union_pred_acc lr lp =
  (match lp, lr with
    [], _ -> lr
  | Pfalse::lp', _ -> [Pfalse]
  | Ptrue::lp', _ -> (union_pred_acc lr lp')
  | e::lp', [Ptrue] -> (union_pred_acc [e] lp')
  | e::lp', _ -> (union_pred_acc (lr@[e]) lp')
  )


(**** OMINUS ****) 
(** apply ominus on each formulas of the list i1_in and cumulate the results **)
and ominus loc i1_in = 
  let phi_cset = List.fold_left (ominus_pred_acc loc) ([],Idset.empty) i1_in 
  in
  begin 
(* begin debug
    (Format.print_string "\n\t(Ominus input ");
    (List.iter (fun p -> begin
      (Format.print_string "\n && ");
      (Util.print_predicate Format.std_formatter p)
    end) i1_in);
    (Format.print_string "\n\t)\n\t(Ominus output ");
    (List.iter (fun p -> begin
      (Format.print_string "\n && ");
      (Util.print_predicate Format.std_formatter p)
    end) (fst phi_cset));
    (Format.print_string "\n\t)\n\t)");
   end debug *)
    phi_cset
  end

(** apply ominus on one formula and cumulate the result to phi_cset **)
and ominus_pred_acc loc phi_cset phi =
  match phi_cset, phi with
    ([Pfalse],_), _ | _, Pfalse -> ([Pfalse],Idset.empty)
  | _, _ ->
  (let phi', cset' = ominus_pred loc (Idmap.empty) phi in 
    match phi' with
      [Pfalse] -> ([Pfalse],Idset.empty)
    | _ -> ((fst phi_cset)@(phi'), (Idset.union (snd phi_cset) cset'))
  )

(** apply substitution sigma followed by ominus on @param phi **)
and ominus_pred loc sigma phi = 
  match phi with
    Pvar _  ->  (* Warning: predicate variables not used in CML *)
      [phi], Idset.empty

  | Papp(p,lterm,lty) -> 
      (* case 1: equality of tokens t1=t2 *)
      if ((String.compare (Ident.string p) "eq_token") == 0)
      then (* lterm = [(Tvar x);(Tvar y)] *)
	(
        let t1, _ = ominus_term loc sigma (List.hd lterm)
	and t2, _ = ominus_term loc sigma (List.hd (List.tl lterm)) 
	in (* extract identifiers of tokens in terms *) 
	match (t1,t2) with
	  (Tvar x, Tvar y) -> 
	    if x = y then (* TODO: verify structural equality *)
	      [Ptrue], Idset.empty
	    else if not(Idmap.mem x loc) && not(Idmap.mem y loc) then 
	      [phi], Idset.empty
	    else
	      [Pfalse], Idset.empty
	| (_,_) -> (* Error: not possible after type checking *)
	    raise (ErrorPrePost "t1=t2 formula: result unexpected")
	)
      else (* case 2: place predicate p(t) *)
	if (Idset.mem p (!cpn_model.places)) 
	then (* lterm = [(Tvar t)] *)
	  (
           let t1, _ = ominus_term loc sigma (List.hd lterm)
	   in 
	   match t1 with
	     Tvar x -> 
               (try
                 let pl = Idmap.find x loc in
                 if p = pl then (* TODO: verify the use of structural equality here *)
                   [Ptrue], Idset.empty
                 else 
                   [Pfalse], Idset.empty
	       with Not_found -> (* x is not in loc *)
                 [phi], Idset.empty)
	   | _ -> (* Error: not possible after type checking *)
	        raise (ErrorPrePost "P(t) formula: result unexpected")
	)
      else (* case 3: predicate on colors *)
        let lterm', cset' = List.fold_left
                              (ominus_term_acc loc sigma) 
                              ([],Idset.empty) lterm 
        in  [Papp(p,lterm',lty)], cset'
 
  | Ptrue -> [Ptrue], Idset.empty

  | Pfalse -> [Pfalse], Idset.empty

  | Pimplies (wp,p1,p2) -> 
      let lp1, cset1 = ominus_pred loc sigma p1 in
      (match lp1 with
         [Pfalse] -> [Ptrue], Idset.empty
       | [Ptrue] -> ominus_pred loc sigma p2
       | [p1'] -> 
           let lp2, cset2 = ominus_pred loc sigma p2 in 
           (match lp2 with
             [Ptrue] -> [Ptrue], Idset.empty
           | [p2'] ->
	       [Pimplies (wp, p1', p2')], (Idset.union cset1 cset2)
           | _ -> raise (ErrorPrePost "result unexpected")
           )
       | _ -> raise (ErrorPrePost "result unexpected")
       ) 

  | Pif _ -> raise (ErrorPrePost "If predicate: not yet implemented!")

  | Pand (wp,sym,p1,p2) -> 
      let lp1, cset1 = ominus_pred loc sigma p1 in
      (match lp1 with
         [Pfalse] -> [Pfalse], Idset.empty
       | [Ptrue] -> ominus_pred loc sigma p2
       | [p1'] -> 
           let lp2, cset2 = ominus_pred loc sigma p2 in 
           (match lp2 with
             [Pfalse] -> [Pfalse], Idset.empty
           | [Ptrue] -> [p1'], cset1
           | [p2'] ->
	       [Pand (wp, sym, p1', p2')], (Idset.union cset1 cset2)
           | _ -> raise (ErrorPrePost "result unexpected")
           )
       | _ -> raise (ErrorPrePost "result unexpected")
       ) 
	
  | Por (p1, p2) ->
      let lp1, cset1 = ominus_pred loc sigma p1 in
      (match lp1 with
        [Ptrue] -> [Ptrue], Idset.empty
      | [Pfalse] -> ominus_pred loc sigma p2
      | [p1'] -> 
          let lp2, cset2 = ominus_pred loc sigma p2 in 
          (match lp2 with
            [Ptrue] -> [Ptrue], Idset.empty
          | [Pfalse] -> [p1'], cset1
          | [p2'] ->
              [Por (p1', p2')], (Idset.union cset1 cset2)
          | _ -> raise (ErrorPrePost "result unexpected")
          )
      | _ -> raise (ErrorPrePost "result unexpected")
      ) 

  | Piff (p1,p2) ->
      let lp1, cset1 = ominus_pred loc sigma p1 
      and lp2, cset2 = ominus_pred loc sigma p2 in
      (match lp1, lp2 with
        [Ptrue], [Ptrue] 
      | [Pfalse], [Pfalse] -> [Ptrue], Idset.empty
      | [Ptrue], [Pfalse] 
      | [Pfalse], [Ptrue] -> [Pfalse], Idset.empty
	    (* TODO: more simplifications *)
      | [p1'], [p2'] ->  [Piff (p1', p2')], (Idset.union cset1 cset2)
      | _ -> raise (ErrorPrePost "result unexpected")
      ) 

  | Pnot p -> 
      let lp, cset =  ominus_pred loc sigma p in
      (match lp with
        [Ptrue] -> [Pfalse], Idset.empty
      | [Pfalse] -> [Ptrue], Idset.empty
      | [p1'] -> [Pnot p1'], cset
      | _ -> raise (ErrorPrePost "result unexpected")
      ) 
              
  | Forall (wp,id,id',ty,trig,p) -> 
      let lp, cset = ominus_pred loc sigma p in 
      if (Cpn.is_token_type ty) 
      then (* quantifier over tokens, add formulas *)
        let pl_id = Cpn.decorate_token_triggers id trig in
        let lp_id, cset_id = ominus_pred_and loc sigma id' pl_id p in
        (match lp_id with
          [Pfalse] -> [Pfalse], Idset.empty
        | [Ptrue] -> [Forall(wp,id,id',ty,trig,(List.hd lp))], cset
        | [p1'] -> [Pand(false,true,Forall(wp,id,id',ty,trig,(List.hd lp)), p1')], 
                   (Idset.union cset cset_id)
        | _ -> raise (ErrorPrePost "result unexpected")
        ) 

      else (* quantifier over colors, continue *)
        [Forall (wp,id,id',ty,trig,(List.hd lp))], cset
      
  | Forallb _ ->
      raise (ErrorPrePost "Forallb formulas: not yet implemented!")

  | Exists (id,id',ty,p) ->
      let lp, cset = ominus_pred loc sigma p in 
      if (Cpn.is_token_type ty) 
      then (* quantifier over tokens, add formulas *)
        let lp_id, cset_id = ominus_pred_or loc sigma id' Idset.empty p in
        (match lp_id with
          [Ptrue] -> [Ptrue], Idset.empty
        | [Pfalse] -> [Exists(id,id',ty,(List.hd lp))], cset
        | [p1'] -> [Por(Exists(id,id',ty,(List.hd lp)), p1')], 
                   (Idset.union cset cset_id)
         | _ -> raise (ErrorPrePost "result unexpected")
        ) 

      else (* quantifier over colors, continue *)
        [Exists (id,id',ty,(List.hd lp))], cset

  | Pfpi _  -> 
      raise (ErrorPrePost "Fpi formulas: not yet implemented!") 

  | Pnamed _ ->
      raise (ErrorPrePost "Named formulas: not yet implemented!") 


(** apply sigma and ominus on a list of terms and return the set of colors introduced **)
and ominus_term_acc loc sigma terms_cset term =
  let t,cset = ominus_term loc sigma term in
    (fst terms_cset)@[t], (Idset.union (snd terms_cset) cset)


(** apply sigma and ominus on a term and return the colors introduced **)
and ominus_term loc sigma term =
  match term with
    Tconst _ -> term, Idset.empty
  | Tvar x -> 
      (* apply substitution sigma *)
      (try
        Tvar (Idmap.find x sigma), Idset.empty
      with Not_found ->
        Tvar x, Idset.empty)

  | Tderef x -> (* TODO *) Tderef x, Idset.empty
   
  | Tapp (f,lt,i) -> 
      (* test that f is a color symbol *)
      if (Idset.mem f (!cpn_model.colors)) 
      then (* lt = [(Tvar t)] with t token by typing *)
	(
         let t1, _ = ominus_term loc sigma (List.hd lt) in 
	 match t1 with
	   Tvar x -> 
             if (Idmap.mem x loc)
             then (* substitute by a color variables and add it in cset *)
               let cstring = (Ident.completeString f)^"_"^(Ident.completeString x) in
	       let cid = (Ident.create cstring) in
               (Tvar cid), (Idset.singleton cid)
             else (* nothing *)
               Tapp (f,[t1], i), Idset.empty
           
	 | _ -> (* Warning: not possible after type checking *)
	     term, Idset.empty
	 )
      else (* function on color logic *)
        let lt', cset = List.fold_left
                              (ominus_term_acc loc sigma) 
                              ([],Idset.empty) lt 
        in  Tapp(f,lt',i), cset

(** build the formula /\_{loc} p [loc/x] **)
and ominus_pred_and loc sigma x pl_x phi = 
  Idmap.fold (ominus_pred_and_aux loc sigma x pl_x phi) loc ([Ptrue], Idset.empty)

and ominus_pred_and_aux loc sigma x pl_x phi z pl_z phi_cset_acc =
  if (Idset.is_empty pl_x) || 
     (Idset.mem pl_z pl_x)
  then 
  let lp1, cset1 = ominus_pred loc (Idmap.add x z sigma) phi in
  (match lp1, (fst phi_cset_acc) with
    [Pfalse], _ 
  | _, [Pfalse] -> [Pfalse], Idset.empty
  | [Ptrue], _ -> phi_cset_acc
  | [p1], [Ptrue] -> (lp1, cset1)
  | [p1], [p2] -> [Pand(false,true,p1,p2)], (Idset.union cset1 (snd  phi_cset_acc))
  | _, _ ->
      raise (ErrorPrePost "Forall formulas: unexpected result!")  
  )
  else
  phi_cset_acc


and ominus_pred_or loc sigma id pl_id p = 
  Idmap.fold (ominus_pred_or_aux loc sigma id pl_id p) loc ([Pfalse], Idset.empty)

and ominus_pred_or_aux loc sigma x pl_x phi z pl_z phi_cset_acc =
  if (Idset.is_empty pl_x) || 
     (Idset.mem pl_z pl_x)
  then 
  let lp1, cset1 = ominus_pred loc (Idmap.add x z sigma) phi in
  (match lp1, (fst phi_cset_acc) with
    [Ptrue], _ 
  | _, [Ptrue] -> [Ptrue], Idset.empty
  | [Pfalse], _ -> phi_cset_acc
  | [p1], [Pfalse] -> (lp1, cset1)
  | [p1], [p2] -> [Por(p1,p2)], (Idset.union cset1 (snd  phi_cset_acc))
  | _, _ ->
      raise (ErrorPrePost "Exists formulas: unexpected result!")  
  )
  else
  phi_cset_acc



(**** OPLUS ****)
(** apply oplus on each formulas of the list i1_in and cumulate the results **)
and oplus loc i1_in =  
  List.fold_left (oplus_pred_acc loc) [] i1_in 

(** apply oplus on one formula and cumulate the result to phi_acc **)
and oplus_pred_acc loc phi_acc phi =
  let lp = oplus_pred loc phi in 
  begin
(* begin debug
    (Format.print_string "\n\t(Oplus input ");
    (Util.print_predicate Format.std_formatter phi);
    (Format.print_string "\n\t)\n\t(Oplus output ");
    List.iter (fun p -> begin
      (Format.print_string "\n && ");
      (Util.print_predicate Format.std_formatter p)
    end) lp;
    (Format.print_string "\n\t)\n\t)");
   end debug *)
    phi_acc@lp  
  end

(** apply oplus on @param phi **)
and oplus_pred loc phi = 
  match phi with
    Pvar _  ->  (* Warning: predicate variables not used in CML *)
      [phi]

  | Papp(p,lterm,lty) -> 
      (* case 1: equality of tokens t1=t2 *)
      if ((String.compare (Ident.string p) "eq_token") == 0)
      then (* lterm = [(Tvar x);(Tvar y)] *)
	(
        let t1 = (List.hd lterm)
	and t2 = (List.hd (List.tl lterm)) 
	in (* extract identifiers of tokens in terms *) 
	match (t1,t2) with
	  (Tvar x, Tvar y) -> 
	    if x = y then (* TODO: verify structural equality *) 
	      [Ptrue]
	    else if not(Idmap.mem x loc) && not(Idmap.mem y loc) then 
	      [phi]
	    else
	      [Pfalse]
	| (_,_) -> (* Error: not possible after type checking *)
	    raise (ErrorPrePost "t1=t2 formula: result unexpected")
	)
      else (* case 2: place predicate p(t) *)
	if (Idset.mem p (!cpn_model.places)) 
	then (* lterm = [(Tvar t)] *)
	  (
           let t1 = (List.hd lterm)
	   in 
	   match t1 with
	     Tvar x -> 
               (try
                 let pl = Idmap.find x loc in
                 if p = pl then (* TODO: verify the use of structural equality here *)
                   [Ptrue]
                 else 
                   [Pfalse]
	       with Not_found -> (* x is not in loc *)
                 [phi])
	   | _ -> (* Error: not possible after type checking *)
	        raise (ErrorPrePost "P(t) formula: result unexpected")
	  )
      else (* case 3: predicate on colors *)
        [phi]
 
  | Ptrue -> [Ptrue]

  | Pfalse -> [Pfalse]

  | Pimplies (wp,p1,p2) -> 
      let lp1 = oplus_pred loc p1 in
      (match lp1 with
         [Pfalse] -> [Ptrue]
       | [Ptrue] -> oplus_pred loc p2
       | [p1'] -> 
           let lp2 = oplus_pred loc p2 in 
           (match lp2 with
             [Ptrue] -> [Ptrue]
           | [p2'] ->
	       [Pimplies (wp, p1', p2')]
           | _ -> raise (ErrorPrePost "result unexpected")
           )
       | _ -> raise (ErrorPrePost "result unexpected")
       ) 

  | Pif _ -> raise (ErrorPrePost "If predicate: not yet implemented!")


  | Pand (wp,sym,p1,p2) -> 
      let lp1 = oplus_pred loc p1 in
      (match lp1 with
         [Pfalse] -> [Pfalse]
       | [Ptrue] -> oplus_pred loc p2
       | [p1'] -> 
           let lp2 = oplus_pred loc p2 in 
           (match lp2 with
             [Pfalse] -> [Pfalse]
           | [Ptrue] -> [p1']
           | [p2'] ->
	       [Pand (wp, sym, p1', p2')]
           | _ -> raise (ErrorPrePost "result unexpected")
           )
       | _ -> raise (ErrorPrePost "result unexpected")
       ) 
	
  | Por (p1, p2) ->
      let lp1 = oplus_pred loc p1 in
      (match lp1 with
        [Ptrue] -> [Ptrue]
      | [Pfalse] -> oplus_pred loc p2
      | [p1'] -> 
          let lp2 = oplus_pred loc p2 in 
          (match lp2 with
            [Ptrue] -> [Ptrue]
          | [Pfalse] -> [p1']
          | [p2'] ->
              [Por (p1', p2')]
          | _ -> raise (ErrorPrePost "result unexpected")
          )
      | _ -> raise (ErrorPrePost "result unexpected")
      ) 

  | Piff (p1,p2) ->
      let lp1 = oplus_pred loc p1 
      and lp2 = oplus_pred loc p2 in
      (match lp1, lp2 with
        [Ptrue], [Ptrue] 
      | [Pfalse], [Pfalse] -> [Ptrue]
      | [Ptrue], [Pfalse] 
      | [Pfalse], [Ptrue] -> [Pfalse]
	    (* TODO: more simplifications *)
      | [p1'], [p2'] ->  [Piff (p1', p2')]
      | _ -> raise (ErrorPrePost "result unexpected")
      ) 

  | Pnot p -> 
      let lp = oplus_pred loc p in
      (match lp with
        [Ptrue] -> [Pfalse]
      | [Pfalse] -> [Ptrue]
      | [p1'] -> [Pnot p1']
      | _ -> raise (ErrorPrePost "result unexpected")
      ) 

  | Forall (wp,id,id',ty,trig,p) -> 
      let lp = oplus_pred loc p in 
      if (Cpn.is_token_type ty) 
      then (* quantifier over tokens, add formulas *) (*TODO SELMA: do the same thing for ominus*)
        let pl_id = Cpn.decorate_token_triggers id trig in (*TODO :SELMA may be i have to look into the Pimplies*)
        let lp_id = oplus_pred_or loc id' pl_id in
        (match lp_id with
          [Pfalse] -> [Forall(wp,id,id',ty,trig,(List.hd lp))]
        | [Ptrue] -> [Ptrue]
        | [p1'] -> [Por(Forall(wp,id,id',ty,trig,(List.hd lp)), p1')]
        | _ -> raise (ErrorPrePost "result unexpected")
        ) 

      else (* quantifier over colors, continue *)
        [Forall (wp,id,id',ty,trig,(List.hd lp))]

  | Forallb _ ->
      raise (ErrorPrePost "Forallb formulas: not yet implemented!")

  | Exists (id,id',ty,p) ->
      let lp = oplus_pred loc p in 
      if (Cpn.is_token_type ty) 
      then (* quantifier over tokens, add formulas *)

        (*SELMA first, we collect places over token id in the formula*)  
        let pl_id = collect_places_exist_token id' p Idset.empty in 

        (*second, we compare it with the set of places in loc*) 
        let lp_id = oplus_pred_and loc id' pl_id in

        (*let lp_id = oplus_pred_and loc id' Idset.empty in PB: SET OF PLACES ALWAYS EMPTY*)
        (match lp_id with
          [Ptrue] -> [Exists(id,id',ty,(List.hd lp))]
        | [Pfalse] -> [Pfalse]
        | [p1'] -> [Exists(id,id',ty,Pand(false,true,(List.hd lp), p1'))]
         | _ -> raise (ErrorPrePost "result unexpected")
        ) 

      else (* quantifier over colors, continue *)
        [Exists (id,id',ty,(List.hd lp))]

  | Pfpi _  -> 
      raise (ErrorPrePost "Fpi formulas: not yet implemented!") 

  | Pnamed _ ->
      raise (ErrorPrePost "Named formulas: not yet implemented!") 

(*SELMA: look like the function  Cpn.decorate_subinv_and *)
(*collect the places of the quantified existentially token id in the formula phi*)
(*spl to cummulate the set of places of id, initially empty*)
and collect_places_exist_token id phi spl =
  match phi with
    Pand (_,_,p1,p2) -> 
      let lpl1 = (collect_places_exist_token id p1 spl) in
       let lpl2 = (collect_places_exist_token id p2 spl) in
         Idset.union lpl1 lpl2

     
  | Papp(pl,[(Tvar v)],_) ->
      if (Idset.mem pl (!cpn_model.places) && String.compare (Ident.string id) (Ident.string v) = 0 )
      then (* is a place of the token id*)
        Idset.add pl spl

      else
	 spl

  | _ -> spl


(** build the formula \/_{loc} p [loc/z] **)
and oplus_pred_or loc x pl_x = 
  Idmap.fold (oplus_pred_or_aux loc x pl_x) loc [Pfalse]

and oplus_pred_or_aux loc x pl_x z pl_z phi_acc =
  if (Idset.is_empty pl_x) || 
     (Idset.mem pl_z pl_x)
  then 
  let phi_x_z = Papp((Ident.create "eq_token"), [(Tvar x);(Tvar z)], []) in
        (* TODO: verify the identifier created for eq_token *)
  (match phi_acc with
    [Pfalse] -> [phi_x_z]
  | [Ptrue] -> [Ptrue]
  | [p1] -> [Por(p1,phi_x_z)]
  | _ -> raise (ErrorPrePost "Forall formulas: unexpected result!")  
  )
  else
  phi_acc

and oplus_pred_and loc x pl_x = 
  Idmap.fold (oplus_pred_and_aux loc x pl_x) loc [Ptrue]

and oplus_pred_and_aux loc x pl_x z pl_z phi_acc =
  if (Idset.is_empty pl_x) || 
     (Idset.mem pl_z pl_x)
  then 
  let phi_x_z = Pnot(Papp((Ident.create "eq_token"), [(Tvar x);(Tvar z)], [])) in
        (* TODO: verify the identifier created for eq_token *)
  (match phi_acc with
    [Pfalse] -> [Pfalse]
  | [Ptrue] -> [phi_x_z]
  | [p1] -> [Pand(false,true,p1,phi_x_z)]
  | _ -> raise (ErrorPrePost "Exists formulas: unexpected result!")  
  )
  else
  phi_acc

