
(* This file is part of the cref analyzer, released under GPL licence.
   Please read the COPYING file in the distribution.

 *)

(*
 * Type for maps from programs points to assertions in the program. 
 *)
module OrderedPC = struct
  type t = Spl_syn.point
  let compare = Equation.compare_point
end

module PointMap = Mappe.Make ( OrderedPC )

type map_point_assertion = Cref_ast.assertion PointMap
 
(*
 * Translation from cref programs to spl.
 *
 * Principle of translation:
 *   - only global programs are translated (no procedure/function)
 *   - global variables are declared for NULL, new, and free
 *   - pointer variables are declared as intergers but have special names
 *   - for a pointer variable v are declared 3 pointer variables 
 *         ptr_v, ptr_v_next, ptr_v_data
 *   - ...
 *)
(* test if a type is sllist *)
let is_type_sllist 
    (ty:Types.type_v)
    : bool
  = 
  match ty with
    PureType (PTExternal ([],id)) ->
      if (String.compare (Ident.string pty) ("sllist")) = 0 
      then
	true
      else
	false

  | _ -> false


let translate_var ldecl (id,ty) = 
  if (is_type_sllist ty) = true then
    ldecl @ [((Apron.Var.of_string ("ptr_"^(Ident.string id))),Spl_syn.INT);
	     ((Apron.Var.of_string ("ptr_"^(Ident.string id)^"_next")),Spl_syn.INT);
	     ((Apron.Var.of_string ("ptr_"^(Ident.string id)^"_data")),Spl_syn.INT)]
  else
    ldecl @ [((Apron.Var.of_string ("ptr_"^(Ident.string id))),Spl_syn.INT)]


(* Translate global declarations. *)
let rec translate_global 
    (prog:'a Cref_ast.t)
    : Spl_syn.declaration list * 'a Cref_ast.t_desc
  =
  match prog with
    LGlobal (id,term,p) -> (* TODO: extract typing infos to generate more vars *)
      let decl1 = (Apron.Var.of_string (Ident.string id), Spl_syn.INT),
	  ldecl, prog1 = translate_global p
      in
          (decl1::ldecl, prog1)

  | LSeq (p1, p2) ->
      let l1,p11 = translate_global p1.ldesc,
	  l2,p22= translate_global p2.ldesc
      in
      (l1@l2, LSeq ({ldesc=p11; linfo=p1.linfo}, {ldesc=p22; linfo=p2.linfo}))
  
  | _ -> ([],prog)


(* Translate first function. *)
let translate_fst_fun 
    (prog:'a Cref_ast.t)
    (init:Spl_syn.point)
    : Spl_syn.declaration list * map_point_assertion * 'a Cref_ast.t_desc
  = 
  match prog.ldesc with
    LFun(id, v1_vn, precond, prog1) ->
      (* ignore name of the function *)
      let ldecl = List.fold_left translate_param [] v1_vn
      in
      ldecl, (PointMap.add init precond PointMap.empty), prog1.ldesc

  | _ -> (* ERROR *)
      [], PointMap.empty(), prog

(* Translate local variables and their init to instructions. *)
let rec translate_local 
    (prog:'a Cref_ast.t)
    : Spl_syn.declaration list
  = 
  match prog.ldesc with
    LVar (v,t) -> 
      let ty = prog.linfo.t_result_type in
      translate_var [] (v,ty)
  | LReturn _ 
  | LNew _
  | LFree _
  | LAssign _
  | LAssignFld _ -> []
  | LSeq (p1,p2) ->
      let l1 = translate_local p1,
	l2 = translate_local p2
      in
      l1 @ l2
  | LIf (_,p1,p2) ->
      let l1 = translate_local p1,
	l2 = translate_local p2
      in
      l1 @ l2
  | LWhile (_,_,p1) ->
      translate_local p1
  | LAbsurd -> []
  | LAssertion (_,_,p1) ->
      translate_local p1
  | LPost (p1,_,_) ->
      translate_local p1
  | LAny _ -> []

(* Translate the program into a list of labeled instructions 
 * and a map from labels to conditions.
 *)
let translate_prog 
    (prog:'a Cref_ast.t)
    (lprepost: map_point_assertion)
    :
    Spl_syn.program * map_point_assertion
  = 

(* Main function for translation *)
let translate 
    (prog:'a Cref_ast.t)
    :
    Spl_syn.program * map_point_assertion
  = 
  (* Main Spl procedure: local declarations contain *)
  (*   - ptr_NULL, ptr_FREE, ptr_NEW *)
  let global_cst:Spl_syn.declaration list = 
    [((Apron.Var.of_string "ptr_NULL"),Spl_syn.INT); 
    [((Apron.Var.of_string "ptr_NEW"),Spl_syn.INT); 
    [((Apron.Var.of_string "ptr_FREE"),Spl_syn.INT)] 
  in
  (*   - global variable declarations *)
  let (global_var:Spl_syn.declaration list,
       prog0:'a Cref_ast.t) = translate_global prog 
  in
  (*   - parameters of the first function declaration *)
  let init_point:Spl_syn.point = { line=0;col=0;char=0 }
  in
  let (param_vars:Spl_syn.declaration list,
       lprepost: map_point_assertion,
       prog1:'a Cref_ast.t) =
    translate_fst_fun prog0 init_point
  in
  (*   - local variables declarations 
   *   - ... and statements corresponding to their initial values *)
  let local_vars:Spl_syn.declaration list =
    translate_local prog1
  in
  (* Main program code. *)
  let (prog_spl:Spl_syn.instr list,prog_lass: map_point_assertion) =
    translate_prog prog1 lprepost
  in
  (* Build the final program from components *)
  let pspl = { procedures = {
	       pname = "";
	       pinput = [];
	       poutput = [];
	       plocal = (((global_cst@global_vars)@param_vars)@local_vars);
	       pcode = { bpoint = init_point;
			 instrs = init_spl@prog_spl
		       }
	     }
  in
  (* Display *)
  let _ =
    Cref_options.if_debug_3
      printf "@[<v>%a@]@."
        (PSpl_syn.print_program
          begin fun fmt (point:Spl_syn.point) -> ()
          end)
        pspl
  in 
  (pspl,lass)


					 
