open Logic
open Types
open Ast
open Env
open Pp
open Util
open Cref_options

(**************************************************************************)

(*s Abstract syntax of imperative programs. *)

exception Cref_bad_program of string

type variable = Ast.variable

type label = Ast.label

type assertion = Ast.assertion

type precondition = Ast.precondition

type postcondition = Ast.postcondition

type assert_kind = Ast.assert_kind

type pterm =
  | Lconst of constant
  | Lvar of Ident.t
  | Lderef of Ident.t
  | Lapp of Ident.t * pterm list 
  | Larr of Ident.t * pterm
  | Lfld of Ident.t * Ident.t

(* ['a] is the type of information associated to the nodes. 
   It will be defined later in module [Env] *)
type 'a t = 
  { ldesc : 'a t_desc;
    linfo : 'a }

and 'a t_desc =
  (* declarations *)
  | LGlobal of variable * pterm * 'a t (* from global LetRef *)
  | LFun of variable * type_v binder list * precondition list * 'a t
  | LVar of variable * pterm 
  (* statements *)
  | LReturn of pterm (* pure terms including !x *)
  | LNew of variable * pterm (* new statement *)
  | LFree of variable
  | LAssign of Ident.t * pterm (* variable assignement *)
  | LAssignArr of Ident.t * pterm * pterm (* array element assignement *)
  | LAssignFld of Ident.t * Ident.t * pterm (* field assignement *)
  | LSeq of 'a t * 'a t
  | LIf of pterm * 'a t * 'a t
  | LWhile of assertion option * pterm * 'a t (* while loop *)
  | LAbsurd (* error *)
  (* assertion *)
  | LAssertion of assert_kind * assertion list * 'a t
  | LPost of 'a t * postcondition * transp 
  (* undeterministic expression *)
  | LAny of type_c

(**************************************************************************)

(*s Printing ast. *)

let rec print fmt p =
  match p.ldesc with
  (* declarations *)
  | LGlobal (id, t, p1) -> 
      begin
      (Format.fprintf fmt "Global decl %a @." Ident.print id);
      (print fmt p1)
      end
  | LFun (id,param,pre,p1) -> 
      begin
      (Format.fprintf fmt "fun %a (..)\n" Ident.print id);
      (*(print_list simple_comma print_type_v fmt param); *)
      (Format.fprintf fmt "/* precondition ");
      (print_pre print_assertion fmt pre);
      (Format.fprintf fmt "\n */ {\n");
      (print fmt p1);
      (Format.fprintf fmt "}\n")
      end
  | LVar (id, t) ->
      (Format.fprintf fmt "var %a := %a @. " Ident.print id print_pterm t)
  (* statements *)
  | LReturn t -> 
      (Format.fprintf fmt "Return %a @." print_pterm t)
  | LNew (v,t) -> 
      begin
        (Format.fprintf fmt "%a = New (" Ident.print v);
        (Format.fprintf fmt "%a) @." print_pterm t)
      end
  | LFree v -> 
      (Format.fprintf fmt "Free (%a) @." Ident.print v)
  | LAssign (v,t) -> 
      begin
        (Format.fprintf fmt "%a := %a\n" Ident.print v print_pterm t)
      end
  | LAssignArr (a,ie,t) -> 
      begin
        (Format.fprintf fmt "%a[%a] := %a @." 
		Ident.print a print_pterm ie print_pterm t)
      end
  | LAssignFld (x,f,t) ->
      (Format.fprintf fmt "%a->%a := %a @." Ident.print x Ident.print f print_pterm t)
  | LSeq (p1, p2) -> 
      begin
        (print fmt p1);
        (Format.fprintf fmt "; @.");
        (print fmt p2) 
      end
  | LIf (t, p1, p2) -> 
      begin
        (Format.fprintf fmt "\nif (%a) { @." print_pterm t);
        (print fmt p1);
        (Format.fprintf fmt "\n} else {\n ");
        (print fmt p2);
        (Format.fprintf fmt "}\n")
      end
  | LWhile (None, t, p1) -> 
      begin
        (Format.fprintf fmt "\nwhile (%a) { @." print_pterm t);
        (print fmt p1);
        (Format.fprintf fmt "}\n")
      end
  | LWhile (Some(phi), t, p1) -> 
      begin
        (Format.fprintf fmt "/* inv %a */\n" print_assertion phi);
        (Format.fprintf fmt "while (%a) { @." print_pterm t);
        (print fmt p1);
        (Format.fprintf fmt "}\n")
      end
  | LAbsurd -> Format.fprintf fmt "Absurd stmt@."
  (* assertion *)
  | LAssertion (k, phi, p1) -> 
      begin
        (Format.fprintf fmt "\nassert()@.");
        (print fmt p1)
      end
  | LPost (p1, phi, t) -> 
      begin
        (print fmt p1);
        (Format.fprintf fmt "/*@ postcondition %a */\n" print_assertion (fst phi) )
      end
  (* undeterministic expression *)
  | LAny _ -> Format.fprintf fmt "Any expression\n"

and print_pterm fmt pt =
  match pt with 
  | Lconst c -> Format.fprintf fmt " const "
  | Lvar id -> Format.fprintf fmt " var(%a) " Ident.print id
  | Lderef id -> Format.fprintf fmt " !%a " Ident.print id
  | Lapp (f, tl) -> 
      begin
        (Format.fprintf fmt " %a(" Ident.print f);
	print_list simple_comma print_pterm fmt tl;
        (Format.fprintf fmt ")")
      end
  | Larr (a, ie) ->
      begin
        (Format.fprintf fmt " %a[" Ident.print a);
	(print_pterm fmt ie);
        (Format.fprintf fmt "]")
      end
  | Lfld (x, f) ->
     (Format.fprintf fmt "%a->%a" Ident.print x Ident.print f)

and print_pre print_assertion fmt l =
  if l <> [] then begin
    Format.fprintf fmt "@[ ";
    print_list simple_comma print_assertion fmt l;
    Format.fprintf fmt " @]"
  end

and print_exn print_assertion fmt (x,c) =
  Format.fprintf fmt "| %a => @[%a@]@," Ident.print x print_assertion c

and print_post print_assertion fmt = function
  | None ->
      ()
  | Some (c,[]) ->
      Format.fprintf fmt "@[ %a @]" print_assertion c
  | Some (c, l) ->
      Format.fprintf fmt "@[ %a@ %a @]" print_assertion c
        (print_list space (print_exn print_assertion)) l



(*s Build local ast for imperative programs. *)

let rec build id p = 
  match p.desc with
  | Lam (bl, pre, p1) -> 
      let lp1 = build_stmt p1 true in
      { ldesc = LFun(id, bl, pre, lp1); linfo = p.info }
  | _ -> raise (Cref_bad_program "Not a program")

and build_stmt p infun = 
  match p.desc with
  | Expression t -> (* TODO: only return? *)
      let pt = build_term t in
      { ldesc = (LReturn pt); linfo = p.info }
  | Var id -> (* TODO: only return? *)
      { ldesc = (LReturn (Lvar id)); linfo = p.info }
  | Seq (p1, p2) ->
      let lp1 = build_stmt p1 infun in
      let lp2 = build_stmt p2 infun in 
      { ldesc = LSeq(lp1,lp2); linfo = p.info }
  | Loop (inv, v, p1) -> (* infinite loop *)
      let lbt,lp1 = build_loop p1 in 
      { ldesc = LWhile(inv, lbt, lp1); linfo = p.info }
  | If (pc, p1, p2) -> 
      let bc = build_stmt_term pc in
      let lp1 = build_stmt p1 infun in
      let lp2 = build_stmt p2 infun in
      { ldesc = LIf(bc, lp1, lp2); linfo = p.info }
  | LetRef (v, p1, p2) -> 
      let t1 = build_stmt_term (remove_label p1) in
      let lp2 = build_stmt p2 infun in
      if infun = true then
         { ldesc = LSeq({ ldesc=LVar (v,t1); linfo = p1.info} ,lp2); linfo = p.info }
      else
         { ldesc = LGlobal(v,t1,lp2); linfo = p.info }
  | LetIn (v, p1, p2) ->  (* may be fun or var *)
      let lp1 = remove_label p1 in
      let lp2 = build_stmt p2 infun in
      if infun = true then
         let t1 = build_stmt_term lp1 in
         { ldesc = LSeq({ ldesc=LVar (v,t1); linfo = p1.info} ,lp2); linfo = p.info }
      else
         (match lp1.desc with 
	  | Expression t ->
              let t1 = build_term t in
              { ldesc = LGlobal(v,t1,lp2); linfo = p.info }
          | Lam (bl, pre, p1) ->
	      let lp1 = build_stmt p1 true in
	      { ldesc = LFun(v,bl,pre,lp1); linfo = p.info }
          | _ -> raise (Cref_bad_program "No local functions")
         )
  | Absurd -> 
      { ldesc = LAbsurd; linfo = p.info }
  | Label (lab, p1) -> 
      begin
        (if_debug_2 Format.eprintf "\n label %s removed \n" lab);
        build_stmt p1 infun 
      end 
  | Assertion (k, [], p1) ->      
      build_stmt p1 infun 
  | Assertion (k, alist, p1) ->      
      let lp1 = build_stmt p1 infun in
      { ldesc = LAssertion (k, alist, lp1); linfo = p.info }
  | Post (p1, cond, t) ->
      let lp1 = build_stmt p1 infun in
      { ldesc = LPost (lp1, cond, t); linfo = p.info }
  | Raise _ -> raise (Cref_bad_program "No exceptions")
  | Try (p1, _) -> (* use to encode while loops *)
      build_stmt p1 true
  | Lam (bl, pre, p1) ->
      if infun = true then
        raise (Cref_bad_program "No local functions")
      else 
        raise (Cref_bad_program "No second order functions")
  | Rec _ -> raise (Cref_bad_program "No recursive functions")
  | AppRef (p1, v, i) -> (* reference *)
      (* raise (Cref_bad_program "No partial applied functions") *)
      begin
      (if_debug_2 Format.eprintf "AppRef %s removed\n" (Ident.string v));
      (* TODO *) build_stmt p1 true
      end
  | AppTerm (p1, t, i) -> (* fun call *)
      let pt = build_term t in
      let funid, ptlist = build_funcall p1 [pt] in
      (match p.info.t_result_type with
      | PureType (PTunit) -> (* search for an assignment *)
	{ ldesc = (build_assign funid ptlist); linfo = p.info } 
      |  _ ->
        { ldesc = LReturn(Lapp(funid,ptlist)); linfo = p.info }
      )
  | Any c -> { ldesc = (LAny c); linfo = p.info } 


and remove_label p =
  (match p.desc with
  | Label (lab, p1) -> remove_label p1
  | _ -> p
  )

and build_loop p = 
  (* extract boolean condition and body of the loop with body p1
   * where p1 = if bc then p11 else raise exn
   *)
  match p.desc with
  | If (pc, p1, p2) ->
      let pp2 = remove_label p2 in
      (match pp2.desc with
      | Raise _ ->
	  let lt = build_stmt_term pc in
	  let lp1 = build_stmt p1 true in 
	  lt, lp1
      | _ -> raise (Cref_bad_program "Incorrect exit from loop")
      )
  | _ -> raise (Cref_bad_program "Incorrect while loop")


 
and build_term t =
  match t with
  | Tconst c -> Lconst c
  | Tvar id -> Lvar id
  | Tderef id -> Lderef id
  | Tapp (id, tlist, inst) -> 
    if id = Ident.access then (* array access *)
      (match tlist with
      | at::idx::_ -> 
	  let lat = build_term_var at in
	  let lidx = build_term idx in 
	  Larr(lat,lidx)
      | _ -> raise (Cref_bad_program "Incorrect access to array")
      )
    else if "f" = (Str.last_chars (Ident.string id) 1) then (* field access *)
      (match tlist with
      | t1::_ -> 
	  let lt1 = build_term_var t1 in
	  Lfld(lt1,id)
      | _ -> raise (Cref_bad_program "Incorrect access to array")
      )
    else
      let ltlist = List.map build_term tlist in
      Lapp(id,ltlist)
  | Tnamed (_, t1) -> build_term t1


and build_stmt_term p =
  match p.desc with
  | Expression t -> build_term t
  | Var id -> Lvar id
  | Label (lab, p1) -> build_stmt_term p1
  | Assertion (k, alist, p1) -> build_stmt_term p1
  | AppTerm (p1, t, i) -> 
    let pt = build_term t in
    let funid, ptlist = build_funcall p1 [pt] in
    begin
      Lapp(funid, ptlist)
    end
  | _ -> raise (Cref_bad_program "Incorrect term") (* TODO *)


and build_term_var t =
  (match t with
  | Tvar id -> id
  | Tderef id -> id
  | _ -> raise (Cref_bad_program "Term is not a variable")
  )

and build_funcall p ptlist =
  (match p.desc with
  | Var id ->
    id, ptlist
  | AppTerm (p1, t, inst) ->
      let pt = build_term t in
      (build_funcall p1 (pt::ptlist))
  | AppRef (p1, v, inst) ->
      (build_funcall p1 ((Lvar v)::ptlist))
  | Assertion (_,_,p1) ->
      (build_funcall p1 ptlist)
  | _ -> raise (Cref_bad_program "Bad function call")
  )

and build_assign id ptlist =
  if id = Ident.ref_set then (* assign a reference *)
     (match ptlist with 
      | (Lvar id)::pt1::_ -> 
        LAssign(id, pt1)
      | _ -> raise (Cref_bad_program "Bad assignment")
     )
  else if id = Ident.array_set then
    (match ptlist with
     | (Lvar a)::pie::pt1::_ -> 
       LAssignArr(a, pie, pt1)
     | _ -> raise (Cref_bad_program "Bad array assign")
    )
  else if "setf_" = (Str.string_before (Ident.string id) 5) then
    let fid = (Str.string_after (Ident.string id) 5) in
    (match ptlist with
     | (Lvar x)::pt1::_ -> 
       LAssignFld(x, (Ident.create fid), pt1)
     | _ -> raise (Cref_bad_program "Bad field assign")
    )
  else
    raise (Cref_bad_program "Unknown assignment")

