(** Generating invariants by fix-point symbolic computation *)

(* This file is part of the cref analyzer, released under GPL licence.
   Please read the COPYING file in the distribution.

 *)

open Ast
open Env
open Format
open Util
open Map
open Cref_ast
open Cref_options
open Sll

(**************************************************************************)
(** Options *)
(**************************************************************************)

(**************************************************************************)
(** Generating invariants *)
(**************************************************************************)

let compute_and_display 
  (prog:Spl_syn.program)
  (lass:Cref2spl.map_point_assertion)
  (fgraph:Equation.graph)
  (manager:'a Apron.Manager.t)
  :
  unit
  =
  let
    (previous:
       (Spl_syn.point, int, 'a Apron.Abstract1.t, unit) Fixpoint.output option 
       ref
    )
    =
    ref None
  in
  let fp = 
    Solving.Forward.compute fgraph ~output:(!previous) 
                            manager ~debug:Cref_options.debug
  in
  (* Display *)
  if_debug_3 printf "@[<v>%a@]@."
    (PSpl_syn.print_program
       begin fun (point:Spl_syn.point) ->
         let abs = PSHGraph.attrvertex fp point in
         printf "@[<hov>%a@ %a@]"
           PSpl_syn.print_point point
           Apron.Abstract1.print abs
       end)
     prog
      ;


let generate 
    (prog:Env.typing_info Cref_ast.t)
    : unit
  =
  (* Provisional: 
   *  Transform the program into 
   *  - an SPL program with pointer variables as integer variables and 
   *  - a list of assertions 
   *)
  let (spl_prog:Spl_syn.program, lass)  = Cref2spl.translate prog in
  (* Display *)
  let _ = 
    Cref_options.if_debug_3
      printf "@[<v>%a@]@."
        (PSpl_syn.print_program
          begin fun fmt (point:Spl_syn.point) ->
            let abs = PSHGraph.attrvertex fp point in
            fprintf fmt "@[<hov>%a@ %a@]"
              PSpl_syn.print_point point
              Apron.Abstract1.print abs
          end)
        spl_prog 
  in
  (* 
   * Converting SPL prog into a forward equation system.
   *)
  let (fgraph:Equation.graph) = Syn2equation.Forward.make spl_prog in
  (* Display *)
  let _ = 
    Cref_options.if_debug_3
      printf "Forward equation graph@   @[<v>%a@]@."
      (PSHGraph.print
        PSpl_syn.print_point
        pp_print_int
        (fun fmt () -> pp_print_string fmt "()")
        Equation.print_transfer
        Equation.print_info)
      fgraph
  in
  (*
   * Choose a domain for the computation/analysis.
   *)
  let (manager:'a Apron.Manager.t) =
    (match Cref_options.domain with
    | Sll ->
        (Sll.manager_alloc())
    | Shapes ->
         (Shapes.manager_alloc())
    )
  in
  compute_and_display prog lass fgraph manager




