
(* This file is part of the cref analyzer, released under GPL licence.
   Please read the COPYING file in the distribution.

 *)

open Ast
open Error
open Ptree
open Types
open Env
open Logic
open Logic_decl
open Misc
open Report
open Typing
open Util
open Cref_options


(*s Processing a program declaration. *)

let interp_cref_program loc id p =
  (* The following code is taken from why/src/main.ml 
   * begin copy *)
  let print_if_debug p x = if_debug_3 Format.eprintf "  @[%a@]@." p x in
  begin
    reset_names ();
    let ploc = p.ploc in
    if_verbose_3 Format.eprintf "*** interpreting program %a@." Ident.print id;

    if_debug Format.eprintf "* typing with effects@.";
    let env = Env.empty_progs () in
    let p = Typing.typef Label.empty env p in
    if effect p <> Effect.bottom then
      raise_located ploc (GlobalWithEffects (id, effect p));
    let c = type_c_of_typing_info [] p.info in
    let c =
      { c with c_post = optpost_app (change_label p.info.t_label "") c.c_post }
    in
    let v = c.c_result_type in
    check_for_not_mutable ploc v;
    Env.add_global id v None;
    print_if_debug print_type_c c;
    (* print_if_debug print_expr p; *)
    (* end copy *)

    (* Local treatment of programs *)
    if_verbose Format.printf "*** build internal structures@.";
    let psimple = Cref_ast.build id p in
    print_if_debug Cref_ast.print psimple;
    if Cref_options.postck then
      begin
	if_verbose Format.printf "*** strongest post-cond checking@.";
	Cref_prepost.check psimple
      end;
    if Cref_options.invgen then
      begin
	if_verbose Format.printf "* invariant generation@.";
	Cref_invgen.generate psimple
      end
  end 


(*s Processing of a declaration. *)

let interp_cref_decl ?(_prelude=false) d = 
  match d with 
    | Program (loc, id, p) ->
	(* Programs are function definitions. *)
	(if_verbose_3 Format.printf "* Program %a is dealt@." Ident.print id);
        if Cref_options.old then
	  Main.interp_decl ~_prelude d
        else 
	begin
	  if Env.is_global id then raise_located p.ploc (Clash id);
          (try interp_cref_program loc id p with Exit -> ())
        end

    | Parameter (loc, ext, ids, v) -> 
	(* Needed for programs using linked structures, i.e., setf, etc. *)
	Main.interp_decl ~_prelude d;
	List.iter (if_verbose_3 Format.printf "* Parameter %a is dealt@." Ident.print) ids

    | Exception (loc, id, v) ->
	(* Needed only for why *)
	Main.interp_decl ~_prelude d;
	(if_verbose_3 Format.eprintf "* Exception %a is not dealt@." Ident.print id)

    | Logic (loc, ext, ids, t0) ->
	(* Needed for predicates and operations on linked structures *)
	Main.interp_decl ~_prelude d;
	List.iter (if_verbose_3 Format.printf "* Logic %a dealt@." Ident.print) ids

    | Predicate_def (loc, id, pl, p) -> 
	Main.interp_decl ~_prelude d;
	(if_verbose_3 Format.printf "* Predicate %a dealt@." Ident.print id)

    | Function_def (loc, id, pl, ty, e) ->
	(if_verbose_3 Format.eprintf "* Function %a is not dealt@." Ident.print id)

    | Inductive_def(loc,id,t,indcases) ->
	(if_verbose_3 Format.eprintf "* Inductive definition %a is not dealt@." Ident.print id)

    | Axiom (loc, id, p) -> 
	(if_verbose_3 Format.eprintf "* Axiom %a is not dealt@." Ident.print id)

    | Goal (loc, id, p) -> 
	Main.interp_decl ~_prelude d

    | TypeDecl (loc, ext, vl, id) ->
	(* Needed to specify record types and their partition *)
	Main.interp_decl ~_prelude d;
	(if_verbose_3 Format.printf "* Type %a dealt@." Ident.print id)
	  


(*s Processing of a channel / a file. *)
  
let deal_cref_channel parsef cin =
  let p = parsef cin in
  if not Options.parse_only then List.iter interp_cref_decl p

let deal_cref_file f =
  Main.reset ();
  let cin = open_in f in 
  deal_cref_channel (Main.why_parser f) cin;
  close_in cin;
  let fwe = Filename.chop_extension f in
  if not (Main.single_file ()) then Main.output (Options.out_file fwe)

let check_cref () =
  begin
    (* call checking tool *)
    Format.printf "- check program \n"; 
    ();
    Format.printf "quit.\n"; 
  end


(* Entry to the analyzer *)

let main_cref () =
  let t0 = Unix.times () in
  Main.load_prelude ();
  if Cref_options.files = [] then begin
    deal_cref_channel (Main.why_parser "standard input") stdin;
    check_cref ();
    Main.output (Options.out_file "out")
  end else begin
    List.iter deal_cref_file Cref_options.files;
    check_cref ();
    if Main.single_file () then 
      let lf = Filename.chop_extension (Misc.last Cref_options.files) in
      Main.output (Options.out_file lf)  
  end;
  if Cref_options.show_time then
    let t1 = Unix.times () in
    Format.printf "Cref execution time : %3.2f@." 
      (t1.Unix.tms_utime -. t0.Unix.tms_utime)


let _ = main_cref ()
